import copy
import sys
from math import pi

import numpy as np
from numpy.linalg import eigh
from scipy.special import gamma
# from scipy.linalg import solve_banded
import ase.units as units
from ase.data import atomic_numbers, atomic_names, chemical_symbols
from ase.utils import seterr

from gpaw.xc import XC
from gpaw.gaunt import gaunt
from gpaw.atom.configurations import configurations
from gpaw.atom.radialgd import (AERadialGridDescriptor,
                                AbinitRadialGridDescriptor)


# Velocity of light in atomic units:
c = 2 * units._hplanck / (units._mu0 * units._c * units._e**2)

# Colors for s, p, d, f, g:
colors = 'krgbycmmmmm'


class GaussianBasis:
    def __init__(self, l, alpha_B, rgd, eps=1.0e-7):
        """Guassian basis set for spherically symmetric atom.

        l: int
            Angular momentum quantum number.
        alpha_B: ndarray
            Exponents.
        rgd: GridDescriptor
            Grid descriptor.
        eps: float
            Cutoff for eigenvalues of overlap matrix."""

        self.l = l
        self.alpha_B = alpha_B
        self.rgd = rgd
        self.eps = eps

        A_BB = np.add.outer(alpha_B, alpha_B)
        M_BB = np.multiply.outer(alpha_B, alpha_B)

        # Overlap matrix:
        S_BB = (2 * M_BB**0.5 / A_BB)**(l + 1.5)

        # Kinetic energy matrix:
        T_BB = 2**(l + 2.5) * M_BB**(0.5 * l + 0.75) / gamma(l + 1.5) * (
            gamma(l + 2.5) * M_BB / A_BB**(l + 2.5) -
            0.5 * (l + 1) * gamma(l + 1.5) / A_BB**(l + 0.5) +
            0.25 * (l + 1) * (2 * l + 1) * gamma(l + 0.5) / A_BB**(l + 0.5))

        # Derivative matrix:
        D_BB = 2**(l + 2.5) * M_BB**(0.5 * l + 0.75) / gamma(l + 1.5) * (
            0.5 * (l + 1) * gamma(l + 1) / A_BB**(l + 1) -
            gamma(l + 2) * alpha_B / A_BB**(l + 2))

        # 1/r matrix:
        K_BB = 2**(l + 2.5) * M_BB**(0.5 * l + 0.75) / gamma(l + 1.5) * (
            0.5 * gamma(l + 1) / A_BB**(l + 1))

        # Find set of linearly independent functions.
        # We have len(alpha_B) gaussians (index B) and self.nbasis
        # linearly independent functions (index b).
        s_B, U_BB = eigh(S_BB)
        self.nbasis = int((s_B > eps).sum())

        Q_Bb = np.dot(U_BB[:, -self.nbasis:],
                      np.diag(s_B[-self.nbasis:]**-0.5))

        self.T_bb = np.dot(np.dot(Q_Bb.T, T_BB), Q_Bb)
        self.D_bb = np.dot(np.dot(Q_Bb.T, D_BB), Q_Bb)
        self.K_bb = np.dot(np.dot(Q_Bb.T, K_BB), Q_Bb)

        r_g = rgd.r_g
        with seterr(divide='ignore'):
            # Avoid errors in debug mode from division by zero:
            gaussians_Bg = np.exp(-np.outer(alpha_B, r_g**2)) * r_g**l
        prefactors_B = (2 * (2 * alpha_B)**(l + 1.5) / gamma(l + 1.5))**0.5
        self.basis_bg = np.dot(Q_Bb.T, prefactors_B[:, None] * gaussians_Bg)

    def __len__(self):
        return self.nbasis

    def expand(self, C_xb):
        return np.dot(C_xb, self.basis_bg)

    def calculate_potential_matrix(self, vr_g):
        vr2dr_g = vr_g * self.rgd.r_g * self.rgd.dr_g
        V_bb = np.inner(self.basis_bg[:, 1:],
                        self.basis_bg[:, 1:] * vr2dr_g[1:])
        return V_bb


def coefs(rgd, l, vr_g, e, scalar_relativistic=False, Z=None):
    r_g = rgd.r_g

    x0_g = 2 * (e * r_g - vr_g) * r_g
    x1_g = 2 * (l + 1) * r_g / rgd.dr_g + r_g**2 * rgd.d2gdr2()
    x2_g = r_g**2 / rgd.dr_g**2
    x = 1.0

    if scalar_relativistic:
        x = (1 + l + l**2 - (Z / c)**2)**0.5 - l
        r_g = r_g.copy()
        r_g[0] = 1.0
        v_g = vr_g / r_g
        M_g = 1 + (e - v_g) / (2 * c**2)
        kappa_g = (rgd.derivative(vr_g) - v_g) / r_g / (2 * c**2 * M_g)
        x0_g = (2 * M_g * (e * r_g - vr_g) * r_g +
                (l + x - 1) * kappa_g * r_g +
                (l + x) * (l + x - 1) - l * (l + 1))
        x1_g = (2 * (l + x) * r_g / rgd.dr_g +
                r_g**2 * rgd.d2gdr2() +
                r_g**2 * kappa_g / rgd.dr_g)

    cm1_g = x2_g - x1_g / 2
    c0_g = x0_g - 2 * x2_g
    cp1_g = x2_g + x1_g / 2

    return cm1_g, c0_g, cp1_g, x


class Channel:
    def __init__(self, l, s=0, f_n=(), basis=None):
        self.l = l
        self.s = s
        self.basis = basis

        self.C_nb = None                       # eigenvectors
        self.e_n = None                        # eigenvalues
        self.f_n = np.array(f_n, dtype=float)  # occupation numbers
        self.phi_ng = None                     # wave functions

        self.name = 'spdfg'[l]
        self.solve2ok = False

    def solve(self, vr_g):
        """Diagonalize Schrödinger equation in basis set."""
        H_bb = self.basis.calculate_potential_matrix(vr_g)
        H_bb += self.basis.T_bb
        self.e_n, C_bn = eigh(H_bb)
        self.C_nb = C_bn.T
        self.phi_ng = self.basis.expand(self.C_nb[:len(self.f_n)])

    def solve2(self, vr_g, scalar_relativistic=False, Z=None, rgd=None):
        rgd = rgd or self.basis.rgd
        r_g = rgd.r_g
        l = self.l
        u_g = rgd.empty()
        self.solve2ok = True
        for n in range(len(self.f_n)):
            e = self.e_n[n]

            if e > 0.0:
                # Skip orbitals with positive energies
                self.e_n[n] = 42
                self.phi_ng[n] = 0.0
                continue

            # Find classical turning point:
            x = vr_g * r_g + 0.5 * l * (l + 1) - e * r_g**2
            g0 = rgd.round(4.0)
            while x[g0] > 0:
                g0 -= 1

            iter = 0
            ok = False
            while True:
                du1dr, a = self.integrate_outwards(u_g, rgd, vr_g, g0, e,
                                                   scalar_relativistic, Z)
                u1 = u_g[g0]
                du2dr = self.integrate_inwards(u_g, rgd, vr_g, g0, e,
                                               scalar_relativistic, Z)
                u2 = u_g[g0]
                A = du1dr / u1 - du2dr / u2
                u_g[g0:] *= u1 / u2
                norm = rgd.integrate(u_g**2, -2) / (4 * pi)
                u_g /= norm**0.5
                a /= norm**0.5

                nodes = (u_g[:-1] * u_g[1:] < 0).sum()

                if abs(A) < 1e-7 and nodes == n:
                    ok = True
                    break

                if nodes > n:
                    e *= 1.2
                elif nodes < n:
                    e *= 0.8
                else:
                    e += 0.5 * A * u_g[g0]**2
                    if e > 0:
                        break

                iter += 1
                assert iter < 400, (n, l, e)

            if ok:
                self.e_n[n] = e
                self.phi_ng[n, 1:] = u_g[1:] / r_g[1:]
                self.phi_ng[n, 0] = a
            else:
                self.solve2ok = False

    def calculate_density(self, n=None):
        """Calculate density."""
        if n is None:
            n_g = 0.0
            for n, f in enumerate(self.f_n):
                n_g += f * self.calculate_density(n)
        else:
            n_g = self.phi_ng[n]**2 / (4 * pi)
        return n_g

    def calculate_kinetic_energy_density(self, n):
        """Calculate kinetic energy density."""
        phi_g = self.phi_ng[n]
        rgd = self.basis.rgd
        tau_g = rgd.derivative(phi_g)**2 / (8 * pi)
        if self.l > 0:
            tau_g[1:] += (self.l * (self.l + 1) *
                          (phi_g[1:] / rgd.r_g[1:])**2 / (8 * pi))
        return tau_g

    def get_eigenvalue_sum(self):
        f_n = self.f_n
        return np.dot(f_n, self.e_n[:len(f_n)])

    def integrate_outwards(self, u_g, rgd, vr_g, g0, e,
                           scalar_relativistic=False, Z=None, pt_g=None):
        l = self.l
        r_g = rgd.r_g

        cm1_g, c0_g, cp1_g, x = coefs(rgd, l, vr_g, e, scalar_relativistic, Z)

        b_g = rgd.zeros()

        if Z is not None:
            a0 = 1.0
            if scalar_relativistic:
                dadr = 2 * ((l + x - 1) * c**2 / Z - Z) / (1 + 2 * (l + x))
            else:
                dadr = -Z / (l + 1)
            a1 = a0 + dadr * r_g[1]
        else:
            assert not scalar_relativistic
            if pt_g is None:
                a0 = 1.0
            else:
                b_g[1:] = 2 * pt_g[1:] * r_g[1:]**(2 - l)
                a0 = pt_g[1] / r_g[1]**l / (vr_g[1] / r_g[1] - e)
            a1 = a0

        u_g[0] = 0.0
        g = 1
        agm1 = a0
        ag = a1
        while True:
            u_g[g] = ag * r_g[g]**(l + x)
            agp1 = -(agm1 * cm1_g[g] + ag * c0_g[g] + b_g[g]) / cp1_g[g]
            if g == g0:
                break
            g += 1
            agm1 = ag
            ag = agp1

        r = r_g[g0]
        dr = rgd.dr_g[g0]
        da = 0.5 * (agp1 - agm1)
        dudr = (l + x) * r**(l + x - 1) * ag + r**(l + x) * da / dr

        if l - 1 + x < 0:
            phi0 = a0 * (r_g[1] * 0.1)**(l - 1 + x)
        else:
            phi0 = a0 * 0.0**(l - 1 + x)

        return dudr, phi0

    def integrate_inwards(self, u_g, rgd, vr_g, g0, e,
                          scalar_relativistic=False, Z=None, gmax=None):
        l = self.l
        r_g = rgd.r_g

        cm1_g, c0_g, cp1_g, x = coefs(rgd, l, vr_g, e, scalar_relativistic, Z)

        cm1_g[:g0] = 1.0  # prevent division by zero
        c0_g /= -cm1_g
        cp1_g /= -cm1_g

        if gmax is None:
            gmax = len(u_g)

        g = gmax - 2
        agp1 = 1.0
        u_g[gmax - 1] = agp1 * r_g[gmax - 1]**(l + x)
        with np.errstate(over='raise'):
            try:
                ag = np.exp(-(-2 * e)**0.5 * (r_g[gmax - 2] - r_g[gmax - 1]))
            except FloatingPointError:
                ag = 2e50

        while True:
            u_g[g] = ag * r_g[g]**(l + x)
            if ag > 1e50:
                u_g[g:] /= 1e50
                ag = ag / 1e50
                agp1 = agp1 / 1e50
            agm1 = agp1 * cp1_g[g] + ag * c0_g[g]
            if g == g0:
                break
            g -= 1
            agp1 = ag
            ag = agm1

        r = r_g[g]
        dr = rgd.dr_g[g]
        da = 0.5 * (agp1 - agm1)
        dudr = (l + x) * r**(l + x - 1) * ag + r**(l + x) * da / dr

        return dudr


class DiracChannel(Channel):
    def __init__(self, k, f_n, basis):
        l = (abs(2 * k + 1) - 1) // 2
        Channel.__init__(self, l, 0, f_n, basis)
        self.k = k
        self.j = abs(k) - 0.5
        self.c_nb = None  # eigenvectors (small component)

        self.name += '(%d/2)' % (2 * self.j)

    def solve(self, vr_g):
        """Solve Dirac equation in basis set."""
        nb = len(self.basis)
        V_bb = self.basis.calculate_potential_matrix(vr_g)
        H_bb = np.zeros((2 * nb, 2 * nb))
        H_bb[:nb, :nb] = V_bb
        H_bb[nb:, nb:] = V_bb - 2 * c**2 * np.eye(nb)
        H_bb[nb:, :nb] = -c * (-self.basis.D_bb.T + self.k * self.basis.K_bb)
        e_n, C_bn = eigh(H_bb)
        if self.k < 0:
            n0 = nb
        else:
            n0 = nb + 1
        self.e_n = e_n[n0:].copy()
        self.C_nb = C_bn[:nb, n0:].T.copy()  # large component
        self.c_nb = C_bn[nb:, n0:].T.copy()  # small component

    def calculate_density(self, n=None):
        """Calculate density."""
        if n is None:
            n_g = Channel.calculate_density(self)
        else:
            n_g = (self.basis.expand(self.C_nb[n])**2 +
                   self.basis.expand(self.c_nb[n])**2) / (4 * pi)
            if self.basis.l < 0:
                n_g[0] = n_g[1]
        return n_g


class AllElectronAtom:
    def __init__(self, symbol, xc='LDA', spinpol=False, dirac=False,
                 configuration=None,
                 ee_interaction=True,
                 Z=None,
                 log=None):
        """All-electron calculation for spherically symmetric atom.

        symbol: str (or int)
            Chemical symbol (or atomic number).
        xc: str
            Name of XC-functional.
        spinpol: bool
            If true, do spin-polarized calculation.  Default is spin-paired.
        dirac: bool
            Solve Dirac equation instead of Schrödinger equation.
        configuration: list
            Electronic configuration for symbol, format as in
            gpaw.atom.configurations
        ee_interaction: bool
            Use ee_interaction=False to turn off electron-electron interaction.
            Default is True.
        log: stream
            Text output."""

        if isinstance(symbol, int):
            symbol = chemical_symbols[symbol]
        self.symbol = symbol
        self.Z = Z or atomic_numbers[symbol]

        self.nspins = 1 + int(bool(spinpol))

        self.dirac = bool(dirac)

        self.ee_interaction = ee_interaction

        if configuration is not None:
            self.configuration = copy.deepcopy(configuration)
        else:
            self.configuration = None

        self.scalar_relativistic = False

        if isinstance(xc, str):
            self.xc = XC(xc)
        else:
            self.xc = xc

        self.fd = log or sys.stdout

        self.vr_sg = None  # potential * r
        self.n_sg = 0.0  # density
        self.rgd = None  # radial grid descriptor

        # Energies:
        self.ekin = None
        self.eeig = None
        self.eH = None
        self.eZ = None

        self.channels = None

        self.initialize_configuration(self.configuration)

        self.log('Z:              ', self.Z)
        self.log('Name:           ', atomic_names[atomic_numbers[symbol]])
        self.log('Symbol:         ', symbol)
        self.log('XC-functional:  ', self.xc.name)
        self.log('Equation:       ', ['Schrödinger', 'Dirac'][self.dirac])

        self.method = 'Gaussian basis-set'

    def log(self, *args, **kwargs):
        print(file=self.fd, *args, **kwargs)

    def initialize_configuration(self, configuration=None):
        self.f_lsn = {}

        if configuration is None:
            configuration = configurations[self.symbol][1]

        for n, l, f, e in configuration:

            if l not in self.f_lsn:
                self.f_lsn[l] = [[] for s in range(self.nspins)]
            if self.nspins == 1:
                self.f_lsn[l][0].append(f)
            else:
                # Use Hund's rule:
                f0 = min(f, 2 * l + 1)
                self.f_lsn[l][0].append(f0)
                self.f_lsn[l][1].append(f - f0)

        if 0:
            n = 2 + len(self.f_lsn[2][0])
            if self.f_lsn[0][0][n] == 2:
                self.f_lsn[0][0][n] = 1
                self.f_lsn[2][0][n - 3] += 1

    def add(self, n, l, df=+1, s=None):
        """Add (remove) electrons."""
        if s is None:
            if self.nspins == 1:
                s = 0
            else:
                self.add(n, l, 0.5 * df, 0)
                self.add(n, l, 0.5 * df, 1)
                return

        if l not in self.f_lsn:
            self.f_lsn[l] = [[] for x in range(self.nspins)]

        f_n = self.f_lsn[l][s]
        if len(f_n) < n - l:
            f_n.extend([0] * (n - l - len(f_n)))
        f_n[n - l - 1] += df

    def initialize(self, ngpts=2000, rcut=50.0,
                   alpha1=0.01, alpha2=None, ngauss=50,
                   eps=1.0e-7):
        """Initialize basis sets and radial grid.

        ngpts: int
            Number of grid points for radial grid.
        rcut: float
            Cutoff for radial grid.
        alpha1: float
            Smallest exponent for gaussian.
        alpha2: float
            Largest exponent for gaussian.
        ngauss: int
            Number of gaussians.
        eps: float
            Cutoff for eigenvalues of overlap matrix."""

        if alpha2 is None:
            alpha2 = 50.0 * self.Z**2

        if 1:
            # Use grid with r(0)=0, r(1)=a and r(ngpts)=rcut:
            a = 1 / alpha2**0.5 / 20
            b = (rcut - a * ngpts) / (rcut * ngpts)
            b = 1 / round(1 / b)
            self.rgd = AERadialGridDescriptor(a, b, ngpts)
        else:
            from scipy.optimize import root
            rT = self.Z / 137**2
            r1 = rT / 10 / 5
            sol = root(lambda d: r1 / d * (np.exp(d * (ngpts - 1)) - 1) - rcut,
                       0.1)
            d = sol.x[0]
            a = r1 / d
            self.rgd = AbinitRadialGridDescriptor(a, d, ngpts)

        self.log('Grid points:     %d (%.5f, %.5f, %.5f, ..., %.3f, %.3f)' %
                 ((self.rgd.N,) + tuple(self.rgd.r_g[[0, 1, 2, -2, -1]])))

        # Distribute exponents between alpha1 and alpha2:
        alpha_B = alpha1 * (alpha2 / alpha1)**np.linspace(0, 1, ngauss)
        self.log('Exponents:       %d (%.3f, %.3f, ..., %.3f, %.3f)' %
                 ((ngauss,) + tuple(alpha_B[[0, 1, -2, -1]])))

        # Maximum l value:
        lmax = max(self.f_lsn.keys())

        self.channels = []
        nb_l = []
        if not self.dirac:
            for l in range(lmax + 1):
                basis = GaussianBasis(l, alpha_B, self.rgd, eps)
                nb_l.append(len(basis))
                for s in range(self.nspins):
                    self.channels.append(Channel(l, s, self.f_lsn[l][s],
                                                 basis))
        else:
            for K in range(1, lmax + 2):
                leff = (K**2 - (self.Z / c)**2)**0.5 - 1
                basis = GaussianBasis(leff, alpha_B, self.rgd, eps)
                nb_l.append(len(basis))
                for k, l in [(-K, K - 1), (K, K)]:
                    if l > lmax:
                        continue
                    f_n = self.f_lsn[l][0]
                    j = abs(k) - 0.5
                    f_n = (2 * j + 1) / (4 * l + 2) * np.array(f_n)
                    self.channels.append(DiracChannel(k, f_n, basis))

        self.log('Basis functions: %s (%s)' %
                 (', '.join([str(nb) for nb in nb_l]),
                  ', '.join('spdf'[:lmax + 1])))

        self.vr_sg = self.rgd.zeros(self.nspins)
        self.vr_sg[:] = -self.Z

    def solve(self):
        """Diagonalize Schrödinger equation."""
        self.eeig = 0.0
        for channel in self.channels:
            if self.method == 'Gaussian basis-set':
                channel.solve(self.vr_sg[channel.s])
            else:
                channel.solve2(self.vr_sg[channel.s], self.scalar_relativistic,
                               self.Z)

            self.eeig += channel.get_eigenvalue_sum()

    def calculate_density(self):
        """Calculate elctron density and kinetic energy."""
        self.n_sg = self.rgd.zeros(self.nspins)
        for channel in self.channels:
            self.n_sg[channel.s] += channel.calculate_density()

    def calculate_electrostatic_potential(self):
        """Calculate electrostatic potential and energy."""
        n_g = self.n_sg.sum(0)
        if self.ee_interaction:
            self.vHr_g = self.rgd.poisson(n_g)
        else:
            self.vHr_g = self.rgd.zeros()
        self.eH = 0.5 * self.rgd.integrate(n_g * self.vHr_g, -1)
        self.eZ = -self.Z * self.rgd.integrate(n_g, -1)

    def calculate_xc_potential(self):
        self.vxc_sg = self.rgd.zeros(self.nspins)
        self.exc = self.xc.calculate_spherical(self.rgd, self.n_sg,
                                               self.vxc_sg)

    def step(self):
        self.solve()
        self.calculate_density()
        self.calculate_electrostatic_potential()
        self.calculate_xc_potential()
        self.vr_sg = self.vxc_sg * self.rgd.r_g
        self.vr_sg += self.vHr_g
        self.vr_sg -= self.Z
        self.ekin = (self.eeig -
                     self.rgd.integrate((self.vr_sg * self.n_sg).sum(0), -1))

    def run(self, mix=0.4, maxiter=117, dnmax=1e-9):
        if self.channels is None:
            self.initialize()

        if self.dirac:
            equation = 'Dirac'
        elif self.scalar_relativistic:
            equation = 'scalar-relativistic Schrödinger'
        else:
            equation = 'non-relativistic Schrödinger'
        self.log('\nSolving %s equation using %s:' % (equation, self.method))

        dn = self.Z

        vr_old_sg = None
        n_old_sg = None
        for iter in range(maxiter):
            self.log('.', end='')
            self.fd.flush()
            if iter > 0:
                self.vr_sg *= mix
                self.vr_sg += (1 - mix) * vr_old_sg
                dn = self.rgd.integrate(abs(self.n_sg - n_old_sg).sum(0))
                if dn <= dnmax:
                    self.log('\nConverged in', iter, 'steps')
                    break

            vr_old_sg = self.vr_sg
            n_old_sg = self.n_sg
            self.step()

        self.summary()

        if self.method != 'Gaussian basis-set':
            for channel in self.channels:
                assert channel.solve2ok

        if dn > dnmax:
            raise RuntimeError('Did not converge!')

    def refine(self):
        self.method = 'finite difference'
        self.run(dnmax=1e-6, mix=0.14, maxiter=200)

    def summary(self):
        self.write_states()
        self.write_energies()

    def write_states(self):
        self.log('\n state  occupation         eigenvalue          <r>')
        if self.dirac:
            self.log(' nl(j)               [Hartree]        [eV]    [Bohr]')
        else:
            self.log(' nl                  [Hartree]        [eV]    [Bohr]')
        self.log('-----------------------------------------------------')
        states = []
        for ch in self.channels:
            for n, f in enumerate(ch.f_n):
                states.append((ch.e_n[n], n, ch.s, ch))
        states.sort()
        for e, n, s, ch in states:
            name = str(n + ch.l + 1) + ch.name
            if self.nspins == 2:
                name += '(%s)' % '+-'[s]
            n_g = ch.calculate_density(n)
            rave = self.rgd.integrate(n_g, 1)
            self.log(' %-7s  %6.3f %13.6f  %13.5f %6.3f' %
                     (name, ch.f_n[n], e, e * units.Hartree, rave))

    def write_energies(self):
        self.log('\nEnergies:          [Hartree]           [eV]')
        self.log('--------------------------------------------')
        for text, e in [('kinetic      ', self.ekin),
                        ('coulomb (e-e)', self.eH),
                        ('coulomb (e-n)', self.eZ),
                        ('xc           ', self.exc),
                        ('total        ',
                         self.ekin + self.eH + self.eZ + self.exc)]:
            self.log(' %s %+13.6f  %+13.5f' % (text, e, units.Hartree * e))

        self.calculate_exx()
        self.log('\nExact exchange energy: %.6f Hartree, %.5f eV' %
                 (self.exx, self.exx * units.Hartree))

    def get_channel(self, l=None, s=0, k=None):
        if self.dirac:
            for channel in self.channels:
                if channel.k == k:
                    return channel
        else:
            for channel in self.channels:
                if channel.l == l and channel.s == s:
                    return channel
        raise ValueError

    def get_orbital(self, n, l=None, s=0, k=None):
        channel = self.get_channel(l, s, k)
        return channel.basis.expand(channel.C_nb[n])

    def plot_wave_functions(self, rc=4.0, show=True):
        import matplotlib.pyplot as plt
        for ch in self.channels:
            for n in range(len(ch.f_n)):
                fr_g = ch.basis.expand(ch.C_nb[n]) * self.rgd.r_g
                # fr_g = ch.phi_ng[n]
                name = str(n + ch.l + 1) + ch.name
                lw = 2
                if self.nspins == 2:
                    name += '(%s)' % '+-'[ch.s]
                    if ch.s == 1:
                        lw = 1
                if self.dirac and ch.k > 0:
                    lw = 1
                ls = ['-', '--', '-.', ':'][ch.l]
                n_g = ch.calculate_density(n)
                rave = self.rgd.integrate(n_g, 1)
                gave = self.rgd.round(rave)
                fr_g *= np.sign(fr_g[gave])
                plt.plot(self.rgd.r_g, fr_g,
                         ls=ls, lw=lw, color=colors[n + ch.l], label=name)

        plt.legend(loc='best')
        plt.xlabel('r [Bohr]')
        plt.ylabel('$r\\phi(r)$')
        plt.axis(xmin=0, xmax=rc)
        if show:
            plt.show()

    def logarithmic_derivative(self, l, energies, rcut):
        ch = Channel(l)
        gcut = self.rgd.round(rcut)
        u_g = self.rgd.empty()
        logderivs = []
        d0 = 42.0
        offset = 0
        for e in energies:
            dudr = ch.integrate_outwards(u_g, self.rgd, self.vr_sg[0],
                                         gcut, e, self.scalar_relativistic,
                                         self.Z)[0]
            d1 = np.arctan(dudr / u_g[gcut]) / pi + offset
            if d1 > d0:
                offset -= 1
                d1 -= 1
            logderivs.append(d1)
            d0 = d1

        return np.array(logderivs)

    def calculate_exx(self, s=None):
        if s is None:
            self.exx = sum(self.calculate_exx(s)
                           for s in range(self.nspins)) / self.nspins
            return self.exx

        states = []
        lmax = 0
        for ch in self.channels:
            l = ch.l
            for n, phi_g in enumerate(ch.phi_ng):
                f = ch.f_n[n]
                if f > 0 and ch.s == s:
                    states.append((l, f * self.nspins / 2.0 / (2 * l + 1),
                                   phi_g))
                    if l > lmax:
                        lmax = l

        G_LLL = gaunt(lmax)

        exx = 0.0
        j1 = 0
        for l1, f1, phi1_g in states:
            f = 1.0
            for l2, f2, phi2_g in states[j1:]:
                n_g = phi1_g * phi2_g
                for l in range((l1 + l2) % 2, l1 + l2 + 1, 2):
                    G = (G_LLL[l1**2:(l1 + 1)**2,
                               l2**2:(l2 + 1)**2,
                               l**2:(l + 1)**2]**2).sum()
                    vr_g = self.rgd.poisson(n_g, l)
                    e = f * self.rgd.integrate(vr_g * n_g, -1) / 4 / pi
                    exx -= e * G * f1 * f2
                f = 2.0
            j1 += 1

        return exx


class CLICommand:
    """Solve radial equation for an atom.

    Example:

        gpaw atom Li -f PBE -p  # plot wave functions for a lithium atom
    """

    @staticmethod
    def add_arguments(parser):
        add = parser.add_argument
        add('symbol')
        add('-f', '--xc-functional', type=str, default='LDA',
            help='Exchange-Correlation functional ' +
            '(default value LDA)',
            metavar='<XC>')
        add('-a', '--add', metavar='states',
            help='Add electron(s). Use "1s0.5a" to add 0.5 1s ' +
            'electrons to the alpha-spin channel (use "b" for ' +
            'beta-spin).  The number of electrons defaults to ' +
            'one. Examples: "1s", "2p2b", "4f0.1b,3d-0.1a".')
        add('--spin-polarized', action='store_true')
        add('-d', '--dirac', action='store_true',
            help='Solve Dirac equation.')
        add('-p', '--plot', action='store_true')
        add('-e', '--exponents',
            help='Exponents a: exp(-a*r^2).  Use "-e 0.1:20.0:30" ' +
            'to get 30 exponents from 0.1 to 20.0.')
        add('-l', '--logarithmic-derivatives',
            metavar='spdfg,e1:e2:de,radius',
            help='Plot logarithmic derivatives. ' +
            'Example: -l spdf,-1:1:0.05,1.3. ' +
            'Energy range and/or radius can be left out.')
        add('-n', '--ngrid', help='Specify number of grid points.')
        add('-R', '--rcut', help='Radial cutoff.')
        add('-r', '--refine', action='store_true')
        add('-s', '--scalar-relativistic', action='store_true')
        add('--no-ee-interaction', action='store_true',
            help='Turn off electron-electron interaction.')

    @staticmethod
    def run(args):
        main(args)


def parse_ld_str(s, energies=None, r=2.0):
    parts = s.split(',')
    lvalues = ['spdfg'.find(x) for x in parts.pop(0)]
    if parts:
        e1, e2, de = (float(x) for x in parts.pop(0).split(':'))
    else:
        e1, e2, de = energies
    if parts:
        r = float(parts.pop())
    energies = np.linspace(e1, e2, int((e2 - e1) / de) + 1)
    return lvalues, energies, r


def main(args):
    symbol = args.symbol

    nlfs = []
    if args.add:
        for x in args.add.split(','):
            n = int(x[0])
            l = 'spdfg'.find(x[1])
            x = x[2:]
            if x and x[-1] in 'ab':
                s = int(x[-1] == 'b')
                args.spin_polarized = True
                x = x[:-1]
            else:
                s = None
            if x:
                f = float(x)
            else:
                f = 1
            nlfs.append((n, l, f, s))

    aea = AllElectronAtom(symbol,
                          xc=args.xc_functional,
                          spinpol=args.spin_polarized,
                          dirac=args.dirac,
                          ee_interaction=not args.no_ee_interaction)

    kwargs = {}
    if args.exponents:
        parts = args.exponents.split(':')
        kwargs['alpha1'] = float(parts[0])
        if len(parts) > 1:
            kwargs['alpha2'] = float(parts[1])
            if len(parts) > 2:
                kwargs['ngauss'] = int(parts[2])

    for n, l, f, s in nlfs:
        aea.add(n, l, f, s)

    if args.ngrid:
        kwargs['ngpts'] = int(args.ngrid)
    if args.rcut:
        kwargs['rcut'] = float(args.rcut)

    aea.initialize(**kwargs)
    aea.run()

    if args.refine:
        aea.refine()

    if args.scalar_relativistic:
        aea.scalar_relativistic = True
        aea.refine()

    if args.logarithmic_derivatives:
        lvalues, energies, r = parse_ld_str(args.logarithmic_derivatives,
                                            (-1, 1, 0.05))
        import matplotlib.pyplot as plt
        for l in lvalues:
            ld = aea.logarithmic_derivative(l, energies, r)
            plt.plot(energies, ld, colors[l])
        plt.show()

    if args.plot:
        aea.plot_wave_functions()



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

import pytest
import numpy as np
from gpaw.grid_descriptor import GridDescriptor
from gpaw.transformers import Transformer


@pytest.mark.ci
def test_fd_ops_transformations():
    n = 20
    gd = GridDescriptor((n, n, n))
    np.random.seed(8)
    a = gd.empty()
    a[:] = np.random.random(a.shape)

    gd2 = gd.refine()
    b = gd2.zeros()
    for k in [2, 4, 6, 8]:
        inter = Transformer(gd, gd2, k // 2).apply
        inter(a, b)
        assert abs(gd.integrate(a) - gd2.integrate(b)) < 1e-14

    gd2 = gd.coarsen()
    b = gd2.zeros()
    for k in [2, 4, 6, 8]:
        restr = Transformer(gd, gd2, k // 2).apply
        restr(a, b)
        assert abs(gd.integrate(a) - gd2.integrate(b)) < 1e-14

    # complex versions
    a = gd.empty(dtype=complex)
    a.real = np.random.random(a.shape)
    a.imag = np.random.random(a.shape)

    phase = np.ones((3, 2), complex)

    gd2 = gd.refine()
    b = gd2.zeros(dtype=complex)
    for k in [2, 4, 6, 8]:
        inter = Transformer(gd, gd2, k // 2, complex).apply
        inter(a, b, phase)
        assert abs(gd.integrate(a) - gd2.integrate(b)) < 1e-14

    gd2 = gd.coarsen()
    b = gd2.zeros(dtype=complex)
    for k in [2, 4, 6, 8]:
        restr = Transformer(gd, gd2, k // 2, complex).apply
        restr(a, b, phase)
        assert abs(gd.integrate(a) - gd2.integrate(b)) < 1e-14



"""

Implementation of spherical harmonic expansion of screened Coulomb kernel.
Based on

    János G Ángyán et al 2006 J. Phys. A: Math. Gen. 39 8613


"""


import numpy as np
from scipy.special import erfc, comb, factorial2
from math import factorial


def safeerfc(x):
    taylor = (1
              - 2 * x / np.pi**0.5
              + 2 * x**3 / (3 * np.pi**0.5)
              - x**5 / (5 * np.pi)**0.5)
    return np.where(x < 1e-4, taylor, erfc(x))


def Dnk(n, k, Xi):
    # Eq. 28
    if k == 0:
        sum = 0
        for m in range(1, n + 1):
            sum += 2**(-m) * Xi**(-2 * m) / factorial2(2 * n - 2 * m + 1)
        return (safeerfc(Xi)
                + np.exp(-Xi**2)
                / (np.pi**0.5) * 2**(n + 1) * Xi**(2 * n + 1) * sum)
    # Eq. 29
    sum = 0
    for m in range(1, k + 1):
        sum += (comb(m - k - 1, m - 1) * 2**(k - m) * Xi**(2 * (k - m))
                / factorial2(2 * n + 2 * k - 2 * m + 1))

    return (np.exp(-Xi**2)
            * 2**(n + 1) * (2 * n + 1) * Xi**(2 * n + 1)
            / np.pi**0.5 / factorial(k)
            / (2 * n + 2 * k + 1) * sum)


def Phinj(n, j, Xi, xi):
    # Eq. 30
    sum = 0
    for k in range(j):
        sum += Dnk(n, k, Xi) / (Xi**(n + 1)) * xi**(n + 2 * k)
    return sum


def Hn(n, Xi, xi):
    """

    Helper function (Eq. 24)

    """
    return 1 / (2 * (xi * Xi)**(n + 1)) * ((Xi**(2 * n + 1) + xi**(2 * n + 1)) * safeerfc(Xi + xi) - (Xi**(2 * n + 1) - xi**(2 * n + 1)) * safeerfc(Xi - xi))  # noqa: E501


def Fn(n, Xi, xi):
    """

        Helper function (Eq. 22).

        It appears, that the article has a typo, because the summation
        starts at p=1, but correct results require to start at p=0.

    """
    prefactor = 2 / np.pi**0.5
    result = 0.0

    for p in range(0, n + 1):
        result += (-1 / (4 * Xi * xi))**(p + 1) * factorial(n + p) / (factorial(p) * factorial(n - p)) * ((-1)**(n - p) * np.exp(-(xi + Xi)**2) - np.exp(-(xi - Xi)**2))  # noqa: E501
    taylor = np.exp(-Xi**2 - xi**2) * 2**(n + 1) * (3 + 2 * n + 2 * xi**2 * Xi**2) * xi**n * Xi**n / (np.pi**0.5 * factorial2(2 * n + 3))  # noqa: E501

    return np.where((Xi * xi)**(2 * n + 1) < 1e-6, taylor, prefactor * result)


def Phi(n, mu, R, r):
    """

        The official spherical kernel expansion

    """
    Rg = np.maximum.reduce([R, r])
    Rl = np.minimum.reduce([R, r])

    # Scaling as given by Eq. 16 and the text above.
    Xi = mu * Rg
    xi = mu * Rl

    # Eq. 21
    result = Fn(n, Xi, xi) + Hn(n, Xi, xi)
    for m in range(1, n + 1):
        result += (Fn(n - m, Xi, xi)
                   * (Xi**(2 * m) + xi**(2 * m)) / (xi * Xi)**m)

    result = np.where(xi < [1e-3, 1e-2, 1e-1, 1e-1, 1e-1, 1e-1][n],
                      Phinj(n, 2, Xi, xi), result)
    result *= mu

    return result


"""

Implementation of spherical harmonic expansion ends. GPAW spesific stuff
remains below.

"""


class RadialHSE:
    def __init__(self, rgd, omega):
        self.rgd = rgd
        self.omega = omega

        self.r1_gg = np.zeros((rgd.N, rgd.N))
        self.r2_gg = np.zeros((rgd.N, rgd.N))
        self.d_gg = np.zeros((rgd.N, rgd.N))
        r_g = rgd.r_g.copy()
        r_g[0] = r_g[1]  # XXX
        self.r1_gg[:] = r_g[None, :]
        self.r2_gg[:] = r_g[:, None]
        self.d_gg[:] = rgd.dr_g[None, :] * rgd.r_g[None, :]**2 * 4 * np.pi
        self.V_lgg = {}

    def screened_coulomb(self, n_g, l):
        # Buffer different l-values for optimal performance
        if l not in self.V_lgg:
            kernel_gg = np.reshape(Phi(l, self.omega, self.r1_gg.ravel(),
                                       self.r2_gg.ravel()),
                                   self.d_gg.shape) / (2 * l + 1)
            self.V_lgg[l] = self.d_gg * kernel_gg
        vr_g = (self.V_lgg[l] @ n_g) * self.rgd.r_g
        return vr_g

    def screened_coulomb_dv(self, n_g, l):
        return self.screened_coulomb(n_g, l) * self.rgd.r_g * self.rgd.dr_g
    



"""BLACS distributed matrix object."""
from typing import Dict, Tuple
import numpy as np
import scipy.linalg as linalg

import _gpaw
from gpaw import debug
from gpaw.mpi import serial_comm, _Communicator
import gpaw.utilities.blas as blas


_global_blacs_context_store: Dict[Tuple[_Communicator, int, int], int] = {}


def matrix_matrix_multiply(alpha, a, opa, b, opb, beta=0.0, c=None,
                           symmetric=False):
    """BLAS-style matrix-matrix multiplication.

    Will use dgemm/zgemm/dsyrk/zherk/dsyr2k/zher2k as apropriate or the
    equivalent PBLAS functions for distributed matrices.

    The coefficients alpha and beta are of type float.  Matrices a, b and c
    must have same type (float or complex).  The strings opa and opb must be
    'N', 'T', or 'C' .  For opa='N' and opb='N', the operation performed is
    equivalent to::

        c.array[:] =  alpha * np.dot(a.array, b.array) + beta * c.array

    Replace a.array with a.array.T or a.array.T.conj() for opa='T' and 'C'
    respectively (similarly for opb).

    Use symmetric=True if the result matrix is symmetric/hermetian
    (only lower half of c will be evaluated).
    """
    return _matrix(a).multiply(alpha, opa, _matrix(b), opb,
                               beta, c if c is None else _matrix(c),
                               symmetric)


def suggest_blocking(N, ncpus):
    """Suggest blocking of NxN matrix.

    Returns rows, columns, blocksize tuple."""

    nprow = ncpus
    npcol = 1

    # Make npcol and nprow as close to each other as possible
    npcol_try = npcol
    while npcol_try < nprow:
        if ncpus % npcol_try == 0:
            npcol = npcol_try
            nprow = ncpus // npcol
        npcol_try += 1

    assert npcol * nprow == ncpus

    # ScaLAPACK creates trouble if there aren't at least a few whole blocks.
    # Choose block size so that there will always be at least one whole block
    # and at least two blocks in total.
    blocksize = max((N - 2) // max(nprow, npcol), 1)
    # The next commented line would give more whole blocks.
    # blocksize = max(N // max(nprow, npcol) - 2, 1)

    # Use block size that is a power of 2 and at most 64
    blocksize = 2**int(np.log2(blocksize))
    blocksize = max(min(blocksize, 64), 1)

    return nprow, npcol, blocksize


class Matrix:
    def __init__(self, M, N, dtype=None, data=None, dist=None):
        """Matrix object.

        M: int
            Rows.
        N: int
            Columns.
        dtype: type
            Data type (float or complex).
        dist: tuple or None
            BLACS distribution given as (communicator, rows, colums, blocksize)
            tuple.  Default is None meaning no distribution.
        data: ndarray or None.
            Numpy ndarray to use for starage.  By default, a new ndarray
            will be allocated.
            """
        self.shape = (M, N)

        if dtype is None:
            if data is None:
                dtype = float
            else:
                dtype = data.dtype
        self.dtype = np.dtype(dtype)

        dist = dist or ()
        if isinstance(dist, tuple):
            dist = create_distribution(M, N, *dist)
        self.dist = dist

        if data is None:
            self.array = np.empty(dist.shape, self.dtype)
        else:
            self.array = data.reshape(dist.shape)

        self.comm = serial_comm
        self.state = 'everything is fine'

    def __len__(self):
        return self.shape[0]

    def __repr__(self):
        dist = str(self.dist).split('(')[1]
        return 'Matrix({}: {}'.format(self.dtype.name, dist)

    def new(self, dist='inherit'):
        """Create new matrix of same shape and dtype.

        Default is to use same BLACS distribution.  Use dist to use another
        distribution.
        """
        return Matrix(*self.shape, dtype=self.dtype,
                      dist=self.dist if dist == 'inherit' else dist)

    def __setitem__(self, i, x):
        # assert i == slice(None)
        if isinstance(x, np.ndarray):
            1 / 0  # sssssself.array[:] = x
        else:
            x.eval(self)

    def __iadd__(self, x):
        x.eval(self, 1.0)
        return self

    def multiply(self, alpha, opa, b, opb, beta=0.0, out=None,
                 symmetric=False):
        """BLAS-style Matrix-matrix multiplication.

        See matrix_matrix_multipliction() for details.
        """
        dist = self.dist
        if out is None:
            assert beta == 0.0
            if opa == 'N':
                M = self.shape[0]
            else:
                M = self.shape[1]
            if opb == 'N':
                N = b.shape[1]
            else:
                N = b.shape[0]
            out = Matrix(M, N, self.dtype,
                         dist=(dist.comm, dist.rows, dist.columns))
        if dist.comm.size > 1:
            # Special cases that don't need scalapack - most likely also
            # faster:
            if alpha == 1.0 and opa == 'N' and opb == 'N':
                return fastmmm(self, b, out, beta)
            if alpha == 1.0 and beta == 1.0 and opa == 'N' and opb == 'C':
                if symmetric:
                    return fastmmm2(self, b, out)
                else:
                    return fastmmm2notsym(self, b, out)

        dist.multiply(alpha, self, opa, b, opb, beta, out, symmetric)
        return out

    def redist(self, other):
        """Redistribute to other BLACS layout."""
        if self is other:
            return
        d1 = self.dist
        d2 = other.dist
        n1 = d1.rows * d1.columns
        n2 = d2.rows * d2.columns
        if n1 == n2 == 1:
            other.array[:] = self.array
            return

        if n2 == 1 and d1.blocksize is None:
            assert d2.blocksize is None
            comm = d1.comm
            if comm.rank == 0:
                M = len(self)
                m = (M + comm.size - 1) // comm.size
                other.array[:m] = self.array
                for r in range(1, comm.size):
                    m1 = min(r * m, M)
                    m2 = min(m1 + m, M)
                    comm.receive(other.array[m1:m2], r)
            else:
                comm.send(self.array, 0)
            return

        if n1 == 1 and d2.blocksize is None:
            assert d1.blocksize is None
            comm = d1.comm
            if comm.rank == 0:
                M = len(self)
                m = (M + comm.size - 1) // comm.size
                other.array[:] = self.array[:m]
                for r in range(1, comm.size):
                    m1 = min(r * m, M)
                    m2 = min(m1 + m, M)
                    comm.send(self.array[m1:m2], r)
            else:
                comm.receive(other.array, 0)
            return

        c = d1.comm if d1.comm.size > d2.comm.size else d2.comm
        n = max(n1, n2)
        if n < c.size:
            c = c.new_communicator(np.arange(n))
        if c is not None:
            M, N = self.shape
            d1 = create_distribution(M, N, c,
                                     d1.rows, d1.columns, d1.blocksize)
            d2 = create_distribution(M, N, c,
                                     d2.rows, d2.columns, d2.blocksize)
            if n1 == n:
                ctx = d1.desc[1]
            else:
                ctx = d2.desc[1]
            redist(d1, self.array, d2, other.array, ctx)

    def invcholesky(self):
        """Inverse of Cholesky decomposition.

        Only the lower part is used.
        """
        if self.state == 'a sum is needed':
            self.comm.sum(self.array, 0)

        if self.comm.rank == 0:
            if self.dist.comm.size > 1:
                S = self.new(dist=(self.dist.comm, 1, 1))
                self.redist(S)
            else:
                S = self
            if self.dist.comm.rank == 0:
                if debug:
                    S.array[np.triu_indices(S.shape[0], 1)] = 42.0
                L_nn = linalg.cholesky(S.array,
                                       lower=True,
                                       overwrite_a=True,
                                       check_finite=debug)
                S.array[:] = linalg.inv(L_nn,
                                        overwrite_a=True,
                                        check_finite=debug)
            if S is not self:
                S.redist(self)

        if self.comm.size > 1:
            self.comm.broadcast(self.array, 0)
            self.state == 'everything is fine'

    def eigh(self, cc=False, scalapack=(None, 1, 1, None)):
        """Calculate eigenvectors and eigenvalues.

        Matrix must be symmetric/hermitian and stored in lower half.

        cc: bool
            Complex conjugate matrix before finding eigenvalues.
        scalapack: tuple
            BLACS distribution for ScaLapack to use.  Default is to do serial
            diagonalization.
        """
        slcomm, rows, columns, blocksize = scalapack

        if self.state == 'a sum is needed':
            self.comm.sum(self.array, 0)

        slcomm = slcomm or self.dist.comm
        dist = (slcomm, rows, columns, blocksize)

        redist = (rows != self.dist.rows or
                  columns != self.dist.columns or
                  blocksize != self.dist.blocksize)

        if redist:
            H = self.new(dist=dist)
            self.redist(H)
        else:
            assert self.dist.comm.size == slcomm.size
            H = self

        eps = np.empty(H.shape[0])

        if rows * columns == 1:
            if self.comm.rank == 0 and self.dist.comm.rank == 0:
                if cc and H.dtype == complex:
                    np.negative(H.array.imag, H.array.imag)
                if debug:
                    H.array[np.triu_indices(H.shape[0], 1)] = 42.0
                eps[:], H.array.T[:] = linalg.eigh(H.array,
                                                   lower=True,  # ???
                                                   overwrite_a=True,
                                                   check_finite=debug)
            self.dist.comm.broadcast(eps, 0)
        else:
            if slcomm.rank < rows * columns:
                assert cc
                array = H.array.copy()
                info = _gpaw.scalapack_diagonalize_dc(array, H.dist.desc, 'U',
                                                      H.array, eps)
                assert info == 0, info

            # necessary to broadcast eps when some ranks are not used
            # in current scalapack parameter set
            # eg. (2, 1, 2) with 4 processes
            if rows * columns < slcomm.size:
                H.dist.comm.broadcast(eps, 0)

        if redist:
            H.redist(self)

        assert (self.state == 'a sum is needed') == (
            self.comm.size > 1)
        if self.comm.size > 1:
            self.comm.broadcast(self.array, 0)
            self.comm.broadcast(eps, 0)
            self.state == 'everything is fine'

        return eps

    def complex_conjugate(self):
        """Inplace complex conjugation."""
        if self.dtype == complex:
            np.negative(self.array.imag, self.array.imag)


def _matrix(M):
    """Dig out Matrix object from wrapper(s)."""
    if isinstance(M, Matrix):
        return M
    return _matrix(M.matrix)


class NoDistribution:
    comm = serial_comm
    rows = 1
    columns = 1
    blocksize = None

    def __init__(self, M, N):
        self.shape = (M, N)

    def __str__(self):
        return 'NoDistribution({}x{})'.format(*self.shape)

    def global_index(self, n):
        return n

    def multiply(self, alpha, a, opa, b, opb, beta, c, symmetric):
        if symmetric:
            assert opa == 'N'
            assert opb == 'C' or opb == 'T' and a.dtype == float
            if a is b:
                blas.rk(alpha, a.array, beta, c.array)
            else:
                if beta == 1.0 and a.shape[1] == 0:
                    return
                blas.r2k(0.5 * alpha, a.array, b.array, beta, c.array)
        else:
            blas.mmm(alpha, a.array, opa, b.array, opb, beta, c.array)


class BLACSDistribution:
    serial = False

    def __init__(self, M, N, comm, r, c, b):
        self.comm = comm
        self.rows = r
        self.columns = c
        self.blocksize = b

        key = (comm, r, c)
        context = _global_blacs_context_store.get(key)
        if context is None:
            try:
                context = _gpaw.new_blacs_context(comm.get_c_object(),
                                                  c, r, 'R')
            except AttributeError:
                pass
            else:
                _global_blacs_context_store[key] = context

        if b is None:
            if c == 1:
                br = (M + r - 1) // r
                bc = max(1, N)
            elif r == 1:
                br = M
                bc = (N + c - 1) // c
            else:
                raise ValueError('Please specify block size!')
        else:
            br = bc = b

        if context is None:
            assert b is None
            assert c == 1
            n = N
            m = min((comm.rank + 1) * br, M) - min(comm.rank * br, M)
        else:
            n, m = _gpaw.get_blacs_local_shape(context, N, M, bc, br, 0, 0)
        if n < 0 or m < 0:
            n = m = 0
        self.shape = (m, n)
        lld = max(1, n)
        if context is not None:
            self.desc = np.array([1, context, N, M, bc, br, 0, 0, lld],
                                 np.intc)

    def __str__(self):
        return ('BLACSDistribution(global={}, local={}, blocksize={})'
                .format(*('{}x{}'.format(*shape)
                          for shape in [self.desc[3:1:-1],
                                        self.shape,
                                        self.desc[5:3:-1]])))

    def global_index(self, myi):
        return self.comm.rank * int(self.desc[5]) + myi

    def multiply(self, alpha, a, opa, b, opb, beta, c, symmetric):
        if symmetric:
            assert opa == 'N'
            assert opb == 'C' or opb == 'T' and a.dtype == float
            N, K = a.shape
            if a is b:
                _gpaw.pblas_rk(N, K, alpha, a.array,
                               beta, c.array,
                               a.dist.desc, c.dist.desc,
                               'U')
            else:
                _gpaw.pblas_r2k(N, K, 0.5 * alpha, b.array, a.array,
                                beta, c.array,
                                b.dist.desc, a.dist.desc, c.dist.desc,
                                'U')
        else:
            Ka, M = a.shape
            N, Kb = b.shape
            if opa == 'N':
                Ka, M = M, Ka
            if opb == 'N':
                N, Kb = Kb, N
            _gpaw.pblas_gemm(N, M, Ka, alpha, b.array, a.array,
                             beta, c.array,
                             b.dist.desc, a.dist.desc, c.dist.desc,
                             opb, opa)


def redist(dist1, M1, dist2, M2, context):
    _gpaw.scalapack_redist(dist1.desc, dist2.desc,
                           M1, M2,
                           dist1.desc[2], dist1.desc[3],
                           1, 1, 1, 1,  # 1-indexing
                           context, 'G')


def create_distribution(M, N, comm=None, r=1, c=1, b=None):
    if comm is None or comm.size == 1:
        assert r == 1 and abs(c) == 1 or c == 1 and abs(r) == 1
        return NoDistribution(M, N)

    return BLACSDistribution(M, N, comm,
                             r if r != -1 else comm.size,
                             c if c != -1 else comm.size,
                             b)


def fastmmm(m1, m2, m3, beta):
    comm = m1.dist.comm

    buf1 = m2.array

    N = len(m1)
    n = (N + comm.size - 1) // comm.size

    for r in range(comm.size):
        if r == 0:
            buf2 = np.empty((n, buf1.shape[1]), dtype=buf1.dtype)

        rrequest = None
        srequest = None
        if r < comm.size - 1:
            rrank = (comm.rank + r + 1) % comm.size
            rn1 = min(rrank * n, N)
            rn2 = min(rn1 + n, N)
            if rn2 > rn1:
                rrequest = comm.receive(buf2[:rn2 - rn1], rrank, 21, False)
            srank = (comm.rank - r - 1) % comm.size
            if len(m2.array) > 0:
                srequest = comm.send(m2.array, srank, 21, False)

        r0 = (comm.rank + r) % comm.size
        n1 = min(r0 * n, N)
        n2 = min(n1 + n, N)
        blas.mmm(1.0, m1.array[:, n1:n2], 'N', buf1[:n2 - n1], 'N',
                 beta, m3.array)

        beta = 1.0

        if r == 0:
            buf1 = np.empty_like(buf2)

        buf1, buf2 = buf2, buf1

        if rrequest:
            comm.wait(rrequest)
        if srequest:
            comm.wait(srequest)

    return m3


def fastmmm2(a, b, out):
    if a.comm:
        assert b.comm is a.comm
        if a.comm.size > 1:
            assert out.comm == a.comm
            assert out.state == 'a sum is needed'

    comm = a.dist.comm
    M, N = a.shape
    m = (M + comm.size - 1) // comm.size
    mym = len(a.array)

    buf1 = np.empty((m, N), dtype=a.dtype)
    buf2 = np.empty((m, N), dtype=a.dtype)
    half = comm.size // 2
    aa = a.array
    bb = b.array

    for r in range(half + 1):
        rrequest = None
        srequest = None

        if r < half:
            srank = (comm.rank + r + 1) % comm.size
            rrank = (comm.rank - r - 1) % comm.size
            skip = (comm.size % 2 == 0 and r == half - 1)
            m1 = min(rrank * m, M)
            m2 = min(m1 + m, M)
            if not (skip and comm.rank < half) and m2 > m1:
                rrequest = comm.receive(buf1[:m2 - m1], rrank, 11, False)
            if not (skip and comm.rank >= half) and mym > 0:
                srequest = comm.send(b.array, srank, 11, False)

        if not (comm.size % 2 == 0 and r == half and comm.rank < half):
            m1 = min(((comm.rank - r) % comm.size) * m, M)
            m2 = min(m1 + m, M)
            if r == 0:
                # symmmmmmmmmmmmmmmmmmmmmmetricccccccccccccccc
                blas.mmm(1.0, aa, 'N', bb, 'C', 1.0, out.array[:, m1:m2])
            else:
                beta = 1.0 if r <= comm.rank else 0.0
                blas.mmm(1.0, aa, 'N', buf2[:m2 - m1], 'C',
                         beta, out.array[:, m1:m2])
            # out.array[:, m1:m2] = m12.array[:, :m2 - m1]

        if rrequest:
            comm.wait(rrequest)
        if srequest:
            comm.wait(srequest)

        bb = buf1
        buf1, buf2 = buf2, buf1

    requests = []
    blocks = []
    nrows = (comm.size - 1) // 2
    for row in range(nrows):
        for column in range(comm.size - nrows + row, comm.size):
            if comm.rank == row:
                m1 = min(column * m, M)
                m2 = min(m1 + m, M)
                if mym > 0 and m2 > m1:
                    requests.append(
                        comm.send(out.array[:, m1:m2].T.conj().copy(),
                                  column, 12, False))
            elif comm.rank == column:
                m1 = min(row * m, M)
                m2 = min(m1 + m, M)
                if mym > 0 and m2 > m1:
                    block = np.empty((mym, m2 - m1), out.dtype)
                    blocks.append((m1, m2, block))
                    requests.append(comm.receive(block, row, 12, False))

    comm.waitall(requests)
    for m1, m2, block in blocks:
        out.array[:, m1:m2] += block

    return out


def fastmmm2notsym(a, b, out):
    if a.comm:
        assert b.comm is a.comm
        if a.comm.size > 1:
            assert out.comm == a.comm
            assert out.state == 'a sum is needed'

    comm = a.dist.comm
    M, N = a.shape
    m = (M + comm.size - 1) // comm.size
    mym = len(a.array)

    buf1 = np.empty((m, N), dtype=a.dtype)
    buf2 = np.empty((m, N), dtype=a.dtype)
    aa = a.array
    bb = b.array

    for r in range(comm.size):
        rrequest = None
        srequest = None

        if r < comm.size - 1:
            srank = (comm.rank + r + 1) % comm.size
            rrank = (comm.rank - r - 1) % comm.size
            m1 = min(rrank * m, M)
            m2 = min(m1 + m, M)
            if m2 > m1:
                rrequest = comm.receive(buf1[:m2 - m1], rrank, 11, False)
            if mym > 0:
                srequest = comm.send(b.array, srank, 11, False)

        m1 = min(((comm.rank - r) % comm.size) * m, M)
        m2 = min(m1 + m, M)
        # symmmmmmmmmmmmmmmmmmmmmmetricccccccccccccccc ??
        blas.mmm(1.0, aa, 'N', bb[:m2 - m1], 'C', 1.0, out.array[:, m1:m2])

        if rrequest:
            comm.wait(rrequest)
        if srequest:
            comm.wait(srequest)

        bb = buf1
        buf1, buf2 = buf2, buf1

    return out



"""Photoelectron spectra from the (shifted) DOS approach."""

import numpy as np

from gpaw.pes import BasePES


class DOSPES(BasePES):

    """PES derived from density of states with shifted KS-energies.

    """

    def __init__(self, mother, daughter=None, shift=False, vde=None,
                 f_min=0.1):
        self.c_m = mother
        self.c_d = daughter
        self.f = None
        self.be = None
        self.shift = shift
        self.vde = vde
        self.f_min = f_min

    def _calculate(self):
        """Evaluate energies and spectroscopic factors."""

        self.be = []
        self.f = []
        ex_m = []
        for spin in range(self.c_m.get_number_of_spins()):
            # use only the Gamma point
            eps_n = self.c_m.get_eigenvalues(0, spin)
            f_n = self.c_m.get_occupation_numbers(0, spin)
            for e, f in zip(eps_n, f_n):
                if f > self.f_min:
                    self.be.append(-e)
                    self.f.append(f)
                    ex_m.append(-e)
        self.be = np.array(self.be)

        # find HOMO energy
        ex_m.sort()
        e_HOMO = ex_m[0]

        if self.vde is not None:
            assert not self.shift
            energy_shift = -e_HOMO + self.vde
        else:
            if self.shift:
                e_m = self.c_m.get_potential_energy()
                try:
                    energy_shift = float(self.c_d) - e_HOMO
                except (TypeError, AttributeError):
                    e_d = self.c_d.get_potential_energy()
                    energy_shift = e_d - e_m - e_HOMO
            else:
                energy_shift = float(self.shift)

        self.be += energy_shift



import pytest
from ase.parallel import parprint
from gpaw import GPAW


@pytest.mark.later
def test_xc_qna_force(in_tmp_dir, gpw_files):
    calc = GPAW(gpw_files['Cu3Au_qna'], parallel=dict(domain=1))
    atoms = calc.get_atoms()
    # Displace atoms to have non-zero forces in the first place
    atoms[0].position[0] += 0.1

    dx_array = [-0.005, 0.000, 0.005]
    E = []

    for i, dx in enumerate(dx_array):

        atoms[0].position[0] += dx
        atoms.calc = calc
        E.append(atoms.get_potential_energy(force_consistent=True))
        if i == 1:
            F = atoms.get_forces()[0, 0]
        atoms[0].position[0] -= dx

    F_num = -(E[-1] - E[0]) / (dx_array[-1] - dx_array[0])
    F_err = F_num - F

    parprint('Analytical force = ', F)
    parprint('Numerical  force = ', F_num)
    parprint('Difference       = ', F_err)
    assert abs(F_err) < 0.01, F_err
    eerr = abs(E[-1] - 277.546724 + 0.12677918230332352)
    assert eerr < 1e-3, eerr



import numpy as np
from gpaw.xc import XC


class XCFunctional:
    def __init__(self, params: dict, ncomponents: int):
        if isinstance(params, (dict, str)):
            self.xc = XC(params, collinear=(ncomponents < 4))
        else:
            self.xc = params
        self.setup_name = self.xc.get_setup_name()
        self.name = self.xc.name
        self.no_forces = self.name.startswith('GLLB')
        self.type = self.xc.type

    def __str__(self):
        return f'name: {self.xc.get_description()}'

    def calculate(self, nt_sr, vxct_sr) -> float:
        if nt_sr.xp is np:
            vxct_sr.data[:] = 0.0
            return self.xc.calculate(nt_sr.desc._gd, nt_sr.data, vxct_sr.data)
        vxct_np_sr = np.zeros(vxct_sr.data.shape)
        exc = self.xc.calculate(nt_sr.desc._gd, nt_sr.data.get(), vxct_np_sr)
        vxct_sr.data[:] = vxct_sr.xp.asarray(vxct_np_sr)
        return exc

    def calculate_paw_correction(self, setup, d, h=None):
        return self.xc.calculate_paw_correction(setup, d, h)

    def get_setup_name(self):
        return self.name



import ase.db
from ase.build import molecule
from ase.optimize.bfgs import BFGS
from ase.data.g2_1 import molecule_names, atom_names

from gpaw import GPAW, PW, Mixer, FermiDirac


c = ase.db.connect('g2-1.db')

for name in molecule_names + atom_names:
    id = c.reserve(name=name, calculator='gpaw')
    if id is None:
        continue
    atoms = molecule(name)
    atoms.cell = [12, 13.01, 14.02]
    atoms.center()
    atoms.calc = GPAW(mode=PW(800),
                      xc='PBE',
                      occupations=FermiDirac(0.0, fixmagmom=True),
                      txt=name + '.txt')
    if name in ['Na2', 'NaCl', 'NO', 'ClO', 'Cl', 'Si']:
        atoms.calc = atoms.calc.new(eigensolver='dav',
                                    mixer=Mixer(0.05, 2))
    atoms.get_forces()
    c.write(atoms, id=id, name=name, relaxed=False)
    if len(atoms) > 1:
        opt = BFGS(atoms, logfile=name + '.gpaw.log')
        opt.run(0.01)
        c.write(atoms, name=name, relaxed=True)



import pytest
from ase.build import molecule

from gpaw import GPAW, Davidson, MixerSum

# Move atom infinitesimally across cell border and test that SCF loop is still
# well converged afterwards.  If it is /not/ well converged, then the code
# which compensates for discontinuity of phases is probably broken.


@pytest.mark.later
def test_generic_move_across_cell():
    def test(calc):
        atoms = molecule('H2O', vacuum=2.5)
        atoms.pbc = 1

        # Translate O to corner:
        atoms.positions -= atoms.positions[0, None, :]

        # Be sure that we are on the positive axis:

        atoms.calc = calc

        eps = 1e-12
        atoms.positions[0, :] = eps
        atoms.get_potential_energy()
        atoms.positions[0, 2] -= 2 * eps
        atoms.get_potential_energy()

        print(calc.scf.niter)

        # We should be within the convergence criterion.
        # It runs a minimum of three iterations:
        assert calc.scf.niter == 3

    def kwargs():
        # Make sure MixerSum works for spin-paired system also:
        return dict(xc='oldLDA', mixer=MixerSum(0.7), kpts=[1, 1, 2])

    test(GPAW(mode='lcao', basis='sz(dzp)', h=0.3))
    test(GPAW(mode='pw', eigensolver=Davidson(3),
              experimental={'reuse_wfs_method': 'paw'}, **kwargs()))
    test(GPAW(mode='fd', h=0.3,
              experimental={'reuse_wfs_method': 'lcao'}, **kwargs()))

    # pw + lcao extrapolation is currently broken (PWLFC lacks integrate2):
    # test(GPAW(mode='pw',
    #           experimental={'reuse_wfs_method': 'lcao'},
    #           **kwargs()))



import pytest
from gpaw.mpi import world
from ase import Atoms, Atom
from ase.build import molecule
from ase.parallel import barrier
from ase.units import Hartree, mol, kcal
from gpaw import GPAW
from gpaw.mixer import Mixer, MixerSum
from gpaw.occupations import FermiDirac
from gpaw.test import gen
from gpaw.mpi import rank

pytestmark = pytest.mark.skipif(world.size < 4,
                                reason='world.size < 4')


@pytest.mark.slow
def test_exx_AA_enthalpy(in_tmp_dir, add_cwd_to_setup_paths):
    data = {}

    def _xc(name):
        return {'name': name, 'stencil': 1}

    # data (from tables.pdf of 10.1063/1.1626543)
    data['N'] = {
        # intermolecular distance (A),
        # formation enthalpy(298) (kcal/mol) on B3LYP geometry
        'exp': (1.098, 0.0, 'none', 'none'),
        'HCTH407': (1.097, 7.9, 'HCTH407', 'gga'),
        'PBE': (1.103, -15.1, 'PBE', 'gga'),
        'BLYP': (1.103, -11.7, 'BLYP', 'gga'),
        'BP86': (1.104, -15.6, 'BP86', 'gga'),
        'BPW91': (1.103, -8.5, 'BPW91', 'gga'),
        'B3LYP': (1.092, -1.03, 'BLYP', 'hyb_gga'),
        'B3PW91': (1.091, 2.8, 'PW91', 'hyb_gga'),
        'PBE0': (1.090, 3.1, 'PBE', 'hyb_gga'),
        'PBEH': (1.090, 3.1, 'PBE', 'hyb_gga'),
        'magmom': 3.0,
        # tables.pdf:
        # https://aip.scitation.org/doi/suppl/10.1063/1.1626543/suppl_file/tables.pdf
        'R_AA_B3LYP': 1.092,  # (from tables.pdf of 10.1063/1.1626543) (Ang)
        'ZPE_AA_B3LYP': 0.005457 * Hartree,  # (from benchmarks.txt of
                                             # 10.1063/1.1626543) (eV)
        'H_298_H_0_AA_B3LYP': 0.003304 * Hartree,  # (from benchmarks.txt of
                                                   # 10.1063/1.1626543) (eV)
        'H_298_H_0_A': 1.04 / (mol / kcal),  # (from 10.1063/1.473182) (eV)
        'dHf_0_A': 112.53 / (mol / kcal)}  # (from 10.1063/1.473182) (eV)

    data['O'] = {
        # intermolecular distance (A),
        # formation enthalpy(298) (kcal/mol) on B3LYP geometry
        'exp': (1.208, 0.0, 'none', 'none'),
        'HCTH407': (1.202, -14.5, 'HCTH407', 'gga'),
        'PBE': (1.218, -23.6, 'PBE', 'gga'),
        'BLYP': (1.229, -15.4, 'BLYP', 'gga'),
        'BP86': (1.220, -21.9, 'BP86', 'gga'),
        'BPW91': (1.219, -17.9, 'BPW91', 'gga'),
        'B3LYP': (1.204, -3.7, 'BLYP', 'hyb_gga'),
        'B3PW91': (1.197, -5.1, 'PW91', 'hyb_gga'),
        'PBE0': (1.192, -4.3, 'PBE', 'hyb_gga'),
        'PBEH': (1.192, -4.3, 'PBE', 'hyb_gga'),
        'magmom': 2.0,
        # tables.pdf:
        # https://aip.scitation.org/doi/suppl/10.1063/1.1626543/suppl_file/tables.pdf
        'R_AA_B3LYP': 1.204,  # (from tables.pdf of 10.1063/1.1626543) (Ang)
        'ZPE_AA_B3LYP': 0.003736 * Hartree,  # (from benchmarks.txt of
                                             # 10.1063/1.1626543) (eV)
        'H_298_H_0_AA_B3LYP': 0.003307 * Hartree,
        # (from benchmarks.txt of 10.1063/1.1626543) (eV)
        'H_298_H_0_A': 1.04 / (mol / kcal),  # (from 10.1063/1.473182) (eV)
        'dHf_0_A': 58.99 / (mol / kcal)}  # (from 10.1063/1.473182) (eV)

    data['H'] = {
        # intermolecular distance (A),
        # formation enthalpy(298) (kcal/mol) on B3LYP geometry
        'exp': (0.741, 0.0, 'none', 'none'),
        'HCTH407': (0.744, 1.8, 'HCTH407', 'gga'),
        'PBE': (0.750, 5.1, 'PBE', 'gga'),
        'BLYP': (0.746, 0.3, 'BLYP', 'gga'),
        'BP86': (0.750, -1.8, 'BP86', 'gga'),
        'BPW91': (0.748, 4.0, 'BPW91', 'gga'),
        'B3LYP': (0.742, -0.5, 'BLYP', 'hyb_gga'),
        'B3PW91': (0.744, 2.4, 'PW91', 'hyb_gga'),
        'PBE0': (0.745, 5.3, 'PBE', 'hyb_gga'),
        'PBEH': (0.745, 5.3, 'PBE', 'hyb_gga'),
        'magmom': 1.0,
        # tables.pdf:
        # https://aip.scitation.org/doi/suppl/10.1063/1.1626543/suppl_file/tables.pdf
        'R_AA_B3LYP': 0.742,  # (from tables.pdf of 10.1063/1.1626543) (Ang)
        'ZPE_AA_B3LYP': 0.010025 * Hartree,  # (from benchmarks.txt of
                                             # 10.1063/1.1626543) (eV)
        'H_298_H_0_AA_B3LYP': 0.003305 * Hartree,  # (from benchmarks.txt of
                                                   # 10.1063/1.1626543) (eV)
        'H_298_H_0_A': 1.01 / (mol / kcal),  # (from 10.1063/1.473182) (eV)
        'dHf_0_A': 51.63 / (mol / kcal)}  # (from 10.1063/1.473182) (eV)

    def calculate(element, h, vacuum, xc, magmom):

        atom = Atoms([Atom(element, (0, 0, 0))])
        if magmom > 0.0:
            mms = [magmom for i in range(len(atom))]
            atom.set_initial_magnetic_moments(mms)

        atom.center(vacuum=vacuum)

        mixer = MixerSum(beta=0.4)
        if element == 'O':
            mixer = MixerSum(0.4, nmaxold=1, weight=100)
            atom.set_positions(atom.get_positions() + [0.0, 0.0, 0.0001])

        calc_atom = GPAW(h=h, xc=_xc(data[element][xc][2]),
                         experimental={'niter_fixdensity': 2},
                         eigensolver='rmm-diis',
                         occupations=FermiDirac(0.0, fixmagmom=True),
                         mixer=mixer,
                         parallel=dict(augment_grids=True),
                         nbands=-2,
                         txt='%s.%s.txt' % (element, xc))
        atom.calc = calc_atom

        mixer = Mixer(beta=0.4, weight=100)
        compound = molecule(element + '2')
        if compound == 'O2':
            mixer = MixerSum(beta=0.4)
            mms = [1.0 for i in range(len(compound))]
            compound.set_initial_magnetic_moments(mms)

        calc = GPAW(h=h, xc=_xc(data[element][xc][2]),
                    experimental={'niter_fixdensity': 2},
                    eigensolver='rmm-diis',
                    mixer=mixer,
                    parallel=dict(augment_grids=True),
                    txt='%s2.%s.txt' % (element, xc))
        compound.set_distance(0, 1, data[element]['R_AA_B3LYP'])
        compound.center(vacuum=vacuum)

        compound.calc = calc

        if data[element][xc][3] == 'hyb_gga':  # only for hybrids
            e_atom = atom.get_potential_energy()
            e_compound = compound.get_potential_energy()

            atom.calc = calc_atom.new(xc=_xc(xc))
            compound.calc = calc.new(xc=_xc(xc))

        e_atom = atom.get_potential_energy()
        e_compound = compound.get_potential_energy()

        dHf_0 = (e_compound - 2 * e_atom + data[element]['ZPE_AA_B3LYP'] +
                 2 * data[element]['dHf_0_A'])
        dHf_298 = (dHf_0 + data[element]['H_298_H_0_AA_B3LYP'] -
                   2 * data[element]['H_298_H_0_A']) * (mol / kcal)
        de = dHf_298 - data[element][xc][1]
        E[element][xc] = de
        if rank == 0:
            print((xc, h, vacuum, dHf_298, data[element][xc][1], de,
                   de / data[element][xc][1]))
            if element == 'H':
                assert dHf_298 == pytest.approx(data[element][xc][1], abs=0.25)
            elif element == 'O':
                assert dHf_298 == pytest.approx(data[element][xc][1], abs=7.5)
            else:
                assert dHf_298 == pytest.approx(data[element][xc][1], abs=2.15)
            assert de == pytest.approx(E_ref[element][xc], abs=0.06)

    E = {}

    E_ref = {'H': {'HCTH407': 0.19286893273630645,
                   'B3LYP': -0.11369634560501423,
                   'PBE0': -0.21413764474738262,
                   'PBEH': -0.14147808591211231},
             'N': {'HCTH407': 2.1354017840869268,
                   'B3LYP': 0.63466589919873972,
                   'PBE0': -0.33376468078480226,
                   'PBEH': -0.30365500626180042}}  # svnversion 5599 # -np 4

    for element in ['H']:  # , 'N']:#, 'O']: # oxygen atom fails to converge
        E[element] = {}
        for xc in ['HCTH407', 'PBE0', 'B3LYP']:
            setup = data[element][xc][2]
            if data[element][xc][3] == 'hyb_gga':  # only for hybrids
                exx = True
            else:
                exx = False
            gen(element, exx=exx, xcname=setup, write_xml=True)
            for h in [0.20]:
                for vacuum in [4.5]:
                    calculate(element, h, vacuum, xc, data[element]['magmom'])
            barrier()



import pytest
from ase import Atoms

from gpaw import GPAW, FermiDirac
from gpaw.analyse.simple_stm import SimpleStm
from gpaw.mpi import rank, size
from gpaw.test import equal


@pytest.mark.later
def test_utilities_simple_stm(in_tmp_dir):
    load = False
    txt = '/dev/null'
    txt = '-'

    me = ''
    if size > 1:
        me += 'rank ' + str(rank) + ': '

    BH = Atoms('BH', [[.0, .0, .41], [.0, .0, -1.23]],
               cell=[5, 6, 6.5])
    BH.center()

    f3dname = 'stm3d.cube'

    def testSTM(calc):
        stm = SimpleStm(calc)
        stm.write_3D([1, 0, 0], f3dname)  # single wf
        wf = stm.gd.integrate(stm.ldos)

        if size == 1:  # XXXX might have trouble reading in parallel
            stm2 = SimpleStm(f3dname)
            wf2 = stm2.gd.integrate(stm2.ldos)
            print('Integrals: written, read=', wf, wf2)
            equal(wf, wf2, 2.e-6)

        stm.scan_const_current(0.02, 5)
    #    print eigenvalue_string(calc)
        stm.write_3D(3.1, f3dname)
        wf2 = stm.gd.integrate(stm.ldos)
    #    print "wf2=", wf2
        equal(wf2, 2, 0.12)

        return wf

    # finite system without spin and width
    fname = 'BH-nospin_wfs.gpw'
    if not load:
        BH.set_pbc(False)
        cf = GPAW(nbands=3, h=.3, txt=txt)
        BH.calc = cf
        e1 = BH.get_potential_energy()
        cf.write(fname, 'all')
    else:
        cf = GPAW(fname, txt=txt)
    wf = testSTM(cf)

    # finite system with spin
    fname = 'BH-spin_Sz2_wfs.gpw'
    BH.set_initial_magnetic_moments([1, 1])
    if not load:
        BH.set_pbc(False)
        cf = GPAW(occupations=FermiDirac(0.1, fixmagmom=True),
                  nbands=5,
                  h=0.3,
                  txt=txt)
        BH.calc = cf
        e2 = BH.get_potential_energy()
        cf.write(fname, 'all')
    else:
        cf = GPAW(fname, txt=txt)
    testSTM(cf)

    # periodic system
    if not load:
        BH.set_pbc(True)
        cp = GPAW(spinpol=True, nbands=3, h=.3,
                  kpts=(2, 1, 1), txt=txt, symmetry='off')
        BH.calc = cp
        e3 = BH.get_potential_energy()
        cp.write('BH-4kpts_wfs.gpw', 'all')
    else:
        cp = GPAW('BH-4kpts_wfs.gpw', txt=txt)

    stmp = SimpleStm(cp)

    stmp.write_3D(-4., f3dname)
    print(me + 'Integrals(occ): 2 * wf, bias=', 2 * wf,
          stmp.gd.integrate(stmp.ldos))
    equal(2 * wf, stmp.gd.integrate(stmp.ldos), 0.02)

    stmp.write_3D(+4., f3dname)
    print(me + 'Integrals(unocc): 2 * wf, bias=', end=' ')
    print(2 * wf, stmp.gd.integrate(stmp.ldos))
    equal(2 * wf, stmp.gd.integrate(stmp.ldos), 0.02)

    energy_tolerance = 0.007
    equal(e1, -2.54026, energy_tolerance)
    equal(e2, -1.51101, energy_tolerance)
    equal(e3, -2.83573, energy_tolerance)



import subprocess
import sys
import os

usage = """gpaw sbatch [-0] -- [sbatch options] script.py [script options]
   or: gpaw sbatch [-0] -- [sbatch options] python -m module [module options]
"""


class CLICommand:
    """Submit a GPAW Python script via sbatch.

    
   If a virtual environment is active when submitting, it will be activated for
   the job as well.
   """

    @staticmethod
    def add_arguments(parser):
        parser.usage = usage
        parser.add_argument('-0', '--test', action='store_true',
                            help='Dry run: Print driver script.')
        parser.add_argument('arguments', nargs='*')

    @staticmethod
    def run(args):
        script = '#!/bin/bash -l\n'
        for i, arg in enumerate(args.arguments):
            if arg.endswith('.py'):
                break
        else:
            for i, arg in enumerate(args.arguments):
                if (arg.startswith('python') and
                    len(args.arguments) > i + 1 and
                    args.arguments[i + 1].startswith('-m')):
                    del args.arguments[i]
                    break
            else:
                print('No script.py found!', file=sys.stderr)
                return

        if arg.endswith('.py'):
            for line in open(arg):
                if line.startswith('#SBATCH'):
                    script += line
        script += ('cd $SLURM_SUBMIT_DIR\n')
        venv = os.getenv('VIRTUAL_ENV')
        if venv:
            print('Detected virtual environment:', venv)
            script += f'source {venv}/bin/activate\n'
        script += ('OMP_NUM_THREADS=1 '
                   'mpiexec `echo $GPAW_MPI_OPTIONS` gpaw python {}\n'
                   .format(' '.join(args.arguments[i:])))
        cmd = ['sbatch'] + args.arguments[:i]
        if args.test:
            print('sbatch command:')
            print(' '.join(cmd))
            print('\nscript:')
            print(script, end='')
        else:
            p = subprocess.Popen(cmd, stdin=subprocess.PIPE, env=os.environ)
            p.communicate(script.encode())



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

"""This module defines a density class."""

from math import pi, sqrt

import numpy as np
from ase.units import Bohr

from gpaw import debug
from gpaw.mixer import get_mixer_from_keywords, MixerWrapper
from gpaw.transformers import Transformer
from gpaw.lfc import LFC, BasisFunctions
from gpaw.wavefunctions.lcao import LCAOWaveFunctions
from gpaw.utilities import (unpack2, unpack_atomic_matrices,
                            pack_atomic_matrices)
from gpaw.utilities.partition import AtomPartition
from gpaw.utilities.timing import nulltimer
from gpaw.arraydict import ArrayDict


class CompensationChargeExpansionCoefficients:
    def __init__(self, setups, nspins):
        self.setups = setups
        self.nspins = nspins

    def calculate(self, D_asp):
        """Calculate multipole moments of compensation charges.

        Returns the total compensation charge in units of electron
        charge, so the number will be negative because of the
        dominating contribution from the nuclear charge."""
        atom_partition = D_asp.partition
        shape_a = [(setup.Delta_pL.shape[1],) for setup in self.setups]
        Q_aL = atom_partition.arraydict(shape_a, dtype=float)
        for a, D_sp in D_asp.items():
            setup = self.setups[a]
            Q_L = np.dot(D_sp[:self.nspins].sum(0), setup.Delta_pL)
            Q_L[0] += setup.Delta0
            Q_aL[a] = Q_L
        return Q_aL

    def get_charge(self, Q_aL):
        local_charge = sqrt(4 * pi) * sum(Q_L[0] for Q_L in Q_aL.values())
        return Q_aL.partition.comm.sum(local_charge)


class NullBackgroundCharge:
    charge = 0.0

    def set_grid_descriptor(self, gd):
        pass

    def add_charge_to(self, rhot_g):
        pass

    def add_fourier_space_charge_to(self, pd, rhot_q):
        pass


class Density:
    """Density object.

    Attributes:
     =============== =====================================================
     ``gd``          Grid descriptor for coarse grids.
     ``finegd``      Grid descriptor for fine grids.
     ``interpolate`` Function for interpolating the electron density.
     ``mixer``       ``DensityMixer`` object.
     =============== =====================================================

    Soft and smooth pseudo functions on uniform 3D grids:
     ========== =========================================
     ``nt_sG``  Electron density on the coarse grid.
     ``nt_sg``  Electron density on the fine grid.
     ``nt_g``   Electron density on the fine grid.
     ``rhot_g`` Charge density on the fine grid.
     ``nct_G``  Core electron-density on the coarse grid.
     ========== =========================================
    """

    def __init__(self, gd, finegd, nspins, collinear, charge, redistributor,
                 background_charge=None):
        """Create the Density object."""

        self.gd = gd
        self.finegd = finegd
        self.nspins = nspins
        self.collinear = collinear
        self.charge = float(charge)
        self.redistributor = redistributor
        self.atomdist = None

        self.ncomponents = self.nspins if self.collinear else 1 + 3

        # This can contain e.g. a Jellium background charge
        if background_charge is None:
            background_charge = NullBackgroundCharge()
        background_charge.set_grid_descriptor(self.finegd)
        self.background_charge = background_charge

        self.charge_eps = 1e-7

        self.D_asp = None
        self.Q = CompensationChargeExpansionCoefficients([], self.nspins)
        self.Q_aL = None

        self.nct_G = None
        self.nt_xG = None
        self.rhot_g = None
        self.nt_xg = None
        self.nt_sg = None
        self.nt_vg = None
        self.nt_g = None

        self.atom_partition = None

        self.setups = None
        self.hund = None
        self.magmom_av = None

        self.fixed = False
        # XXX at least one test will fail because None has no 'reset()'
        # So we need DummyMixer I guess
        self.mixer = None
        self.set_mixer(None)

        self.timer = nulltimer
        self.error = None
        self.nct = None
        self.ghat = None
        self.log = None

    @property
    def nt_sG(self):
        return None if self.nt_xG is None else self.nt_xG[:self.nspins]

    @property
    def nt_vG(self):
        return None if self.nt_xG is None else self.nt_xG[self.nspins:]

    def __str__(self):
        s = 'Densities:\n'
        s += '  Coarse grid: {0}*{1}*{2} grid\n'.format(*self.gd.N_c)
        s += '  Fine grid: {0}*{1}*{2} grid\n'.format(*self.finegd.N_c)
        s += '  Total Charge: {0:.6f}'.format(self.charge)
        if self.fixed:
            s += '\n  Fixed'
        return s

    def summary(self, atoms, magmom, log):
        if self.nspins == 1:
            return
        try:
            # XXX This doesn't always work, HGH, SIC, ...
            sc = self.get_spin_contamination(atoms, int(magmom < 0))
            log('Spin contamination: %f electrons' % sc)
        except (TypeError, AttributeError):
            pass

    def initialize(self, setups, timer, magmom_av, hund):
        self.timer = timer
        self.setups = setups
        self.Q = CompensationChargeExpansionCoefficients(setups, self.nspins)
        self.hund = hund
        self.magmom_av = magmom_av

    def reset(self):
        # TODO: reset other parameters?
        self.nt_xG = None

    def set_positions_without_ruining_everything(self, spos_ac,
                                                 atom_partition):
        rank_a = atom_partition.rank_a
        # If both old and new atomic ranks are present, start a blank dict if
        # it previously didn't exist but it will needed for the new atoms.
        if (self.atom_partition is not None and
            self.D_asp is None and (rank_a == self.gd.comm.rank).any()):
            self.update_atomic_density_matrices(
                self.setups.empty_atomic_matrix(self.ncomponents,
                                                self.atom_partition))

        if (self.atom_partition is not None and self.D_asp is not None and
            self.gd.comm.size > 1):
            self.timer.start('Redistribute')
            self.D_asp.redistribute(atom_partition)
            self.timer.stop('Redistribute')

        self.atom_partition = atom_partition
        self.atomdist = self.redistributor.get_atom_distributions(spos_ac)

    def set_positions(self, spos_ac, atom_partition):
        self.set_positions_without_ruining_everything(spos_ac, atom_partition)
        self.nct.set_positions(spos_ac, atom_partition)
        self.ghat.set_positions(spos_ac, atom_partition)
        self.mixer.reset()

        self.nt_xg = None
        self.nt_sg = None
        self.nt_vg = None
        self.nt_g = None
        self.rhot_g = None

    def calculate_pseudo_density(self, wfs):
        """Calculate nt_sG from scratch.

        nt_sG will be equal to nct_G plus the contribution from
        wfs.add_to_density().
        """
        wfs.calculate_density_contribution(self.nt_xG)
        self.nt_sG[:] += self.nct_G

    def update_atomic_density_matrices(self, value):
        if isinstance(value, dict):
            tmp = self.setups.empty_atomic_matrix(self.ncomponents,
                                                  self.atom_partition)
            tmp.update(value)
            value = tmp
        assert isinstance(value, ArrayDict) or value is None, type(value)
        if value is not None:
            value.check_consistency()
        self.D_asp = value

    def update(self, wfs):
        self.timer.start('Density')
        with self.timer('Pseudo density'):
            self.calculate_pseudo_density(wfs)
        with self.timer('Atomic density matrices'):
            wfs.calculate_atomic_density_matrices(self.D_asp)
        with self.timer('Multipole moments'):
            comp_charge, _Q_aL = self.calculate_multipole_moments()

        if isinstance(wfs, LCAOWaveFunctions):
            self.timer.start('Normalize')
            self.normalize(comp_charge)
            self.timer.stop('Normalize')

        self.timer.start('Mix')
        self.mix(comp_charge)
        self.timer.stop('Mix')
        self.timer.stop('Density')

    def normalize(self, comp_charge):
        """Normalize pseudo density."""

        pseudo_charge = self.gd.integrate(self.nt_sG).sum()

        if (pseudo_charge + self.charge + comp_charge -
            self.background_charge.charge != 0):
            if pseudo_charge != 0:
                x = (self.background_charge.charge - self.charge -
                     comp_charge) / pseudo_charge
                self.nt_xG *= x
            else:
                # Use homogeneous background.
                #
                # We have to use the volume per actual grid point,
                # which is not the same as gd.volume as the latter
                # includes ghost points.
                volume = self.gd.get_size_of_global_array().prod() * self.gd.dv
                total_charge = (self.charge + comp_charge -
                                self.background_charge.charge)
                self.nt_sG[:] = -total_charge / volume

    def mix(self, comp_charge):
        assert isinstance(self.mixer, MixerWrapper), self.mixer
        self.error = self.mixer.mix(self.nt_xG, self.D_asp)
        assert self.error is not None, self.mixer

        comp_charge = None
        self.interpolate_pseudo_density(comp_charge)
        self.calculate_pseudo_charge()

    def calculate_multipole_moments(self):
        D_asp = self.atomdist.to_aux(self.D_asp)
        Q_aL = self.Q.calculate(D_asp)
        self.Q_aL = Q_aL
        return self.Q.get_charge(Q_aL), Q_aL

    def get_initial_occupations(self, a):
        # distribute charge on all atoms
        # XXX interaction with background charge may be finicky
        c = (self.charge - self.background_charge.charge) / len(self.setups)
        M_v = self.magmom_av[a]
        M = np.linalg.norm(M_v)
        f_si = self.setups[a].calculate_initial_occupation_numbers(
            M, self.hund, charge=c,
            nspins=self.nspins if self.collinear else 2)

        if self.collinear:
            if M_v[2] < 0:
                f_si = f_si[::-1].copy()
        else:
            f_i = f_si.sum(0)
            fm_i = f_si[0] - f_si[1]
            f_si = np.zeros((4, len(f_i)))
            f_si[0] = f_i
            if M > 0:
                f_si[1:] = M_v[:, np.newaxis] / M * fm_i

        return f_si

    def initialize_from_atomic_densities(self, basis_functions):
        """Initialize D_asp, nt_sG and Q_aL from atomic densities.

        nt_sG is initialized from atomic orbitals, and will
        be constructed with the specified magnetic moments and
        obeying Hund's rules if ``hund`` is true."""

        # XXX does this work with blacs?  What should be distributed?
        # Apparently this doesn't use blacs at all, so it's serial
        # with respect to the blacs distribution.  That means it works
        # but is not particularly efficient (not that this is a time
        # consuming step)

        self.log('Density initialized from atomic densities')

        self.update_atomic_density_matrices(
            self.setups.empty_atomic_matrix(self.ncomponents,
                                            self.atom_partition))

        f_asi = {}
        for a in basis_functions.atom_indices:
            f_asi[a] = self.get_initial_occupations(a)

        # D_asp does not have the same distribution as the basis functions,
        # so we have to loop over atoms separately.
        for a in self.D_asp:
            f_si = f_asi.get(a)
            if f_si is None:
                f_si = self.get_initial_occupations(a)
            self.D_asp[a][:] = self.setups[a].initialize_density_matrix(f_si)

        self.nt_xG = self.gd.zeros(self.ncomponents)
        basis_functions.add_to_density(self.nt_xG, f_asi)
        self.nt_sG[:] += self.nct_G
        self.calculate_normalized_charges_and_mix()

    def initialize_from_wavefunctions(self, wfs):
        """Initialize D_asp, nt_sG and Q_aL from wave functions."""
        self.log('Density initialized from wave functions')
        self.timer.start('Density initialized from wave functions')
        self.nt_xG = self.gd.zeros(self.ncomponents)
        self.calculate_pseudo_density(wfs)
        self.update_atomic_density_matrices(
            self.setups.empty_atomic_matrix(self.ncomponents,
                                            wfs.atom_partition))
        wfs.calculate_atomic_density_matrices(self.D_asp)
        self.calculate_normalized_charges_and_mix()
        self.timer.stop('Density initialized from wave functions')

    def initialize_directly_from_arrays(self, nt_sG, nt_vG, D_asp):
        """Set D_asp and nt_sG directly."""
        self.nt_xG = self.gd.zeros(self.ncomponents)
        self.nt_sG[:] = nt_sG
        if nt_vG is not None:
            self.nt_vG[:] = nt_vG

        self.update_atomic_density_matrices(D_asp)
        D_asp.check_consistency()
        # No calculate multipole moments?  Tests will fail because of
        # improperly initialized mixer

    def calculate_normalized_charges_and_mix(self):
        comp_charge, _Q_aL = self.calculate_multipole_moments()
        self.normalize(comp_charge)
        self.mix(comp_charge)

    def set_mixer(self, mixer):
        if mixer is None:
            mixer = {}
        if isinstance(mixer, dict):
            mixer = get_mixer_from_keywords(self.gd.pbc_c.any(),
                                            self.ncomponents, **mixer)
        if not hasattr(mixer, 'mix'):
            raise ValueError('Not a mixer: %s' % mixer)
        self.mixer = MixerWrapper(mixer, self.ncomponents, self.gd)

    def calculate_magnetic_moments(self):
        magmom_av = np.zeros_like(self.magmom_av)
        magmom_v = np.zeros(3)
        if self.nspins == 2:
            for a, D_sp in self.D_asp.items():
                M_p = D_sp[0] - D_sp[1]
                magmom_av[a, 2] = np.dot(M_p, self.setups[a].N0_p)
                magmom_v[2] += (np.dot(M_p, self.setups[a].Delta_pL[:, 0]) *
                                sqrt(4 * pi))
            self.gd.comm.sum(magmom_av)
            self.gd.comm.sum(magmom_v)
            magmom_v[2] += self.gd.integrate(self.nt_sG[0] - self.nt_sG[1])
        elif not self.collinear:
            for a, D_sp in self.D_asp.items():
                magmom_av[a] = np.dot(D_sp[1:4], self.setups[a].N0_p)
                magmom_v += (np.dot(D_sp[1:4], self.setups[a].Delta_pL[:, 0]) *
                             sqrt(4 * pi))
            # XXXX probably untested code
            self.gd.comm.sum(magmom_av)
            self.gd.comm.sum(magmom_v)
            magmom_v += self.gd.integrate(self.nt_vG)

        return magmom_v, magmom_av

    estimate_magnetic_moments = calculate_magnetic_moments

    def get_correction(self, a, spin):
        """Integrated atomic density correction.

        Get the integrated correction to the pseuso density relative to
        the all-electron density.
        """
        setup = self.setups[a]
        return sqrt(4 * pi) * (
            np.dot(self.D_asp[a][spin], setup.Delta_pL[:, 0]) +
            setup.Delta0 / self.nspins)

    def get_all_electron_density(self, atoms=None, gridrefinement=2,
                                 spos_ac=None, skip_core=False):
        """Return real all-electron density array.

           Usage: Either get_all_electron_density(atoms) or
                         get_all_electron_density(spos_ac=spos_ac)

           skip_core=True theoretically returns the
                          all-electron valence density (use with
                          care; will not in general integrate
                          to valence)
        """
        if spos_ac is None:
            spos_ac = atoms.get_scaled_positions() % 1.0
        spos_ch = []    # for coreholes

        # Refinement of coarse grid, for representation of the AE-density
        # XXXXXXXXXXXX think about distribution depending on gridrefinement!
        if gridrefinement == 1:
            gd = self.redistributor.aux_gd
            n_sg = self.nt_sG.copy()
            # This will get the density with the same distribution
            # as finegd:
            n_sg = self.redistributor.distribute(n_sg)
        elif gridrefinement == 2:
            gd = self.finegd
            if self.nt_sg is None:
                self.interpolate_pseudo_density()
            n_sg = self.nt_sg.copy()
        elif gridrefinement == 4:
            # Extra fine grid
            gd = self.finegd.refine()

            # Interpolation function for the density:
            interpolator = Transformer(self.finegd, gd, 3)  # XXX grids!

            # Transfer the pseudo-density to the fine grid:
            n_sg = gd.empty(self.nspins)
            if self.nt_sg is None:
                self.interpolate_pseudo_density()
            for s in range(self.nspins):
                interpolator.apply(self.nt_sg[s], n_sg[s])
        else:
            raise NotImplementedError

        if self.nspins > 1 and not skip_core:
            # Support for corehole in spin-polarized system
            spos_ch = []
            n_ch_a = []
            n_ch = []

        # Add corrections to pseudo-density to get the AE-density
        splines = {}
        phi_aj = []
        phit_aj = []
        nc_a = []
        nct_a = []
        for a, id in enumerate(self.setups.id_a):
            if id in splines:
                phi_j, phit_j, nc, nct = splines[id]
            else:
                # Load splines:
                phi_j, phit_j, nc, nct = self.setups[a].get_partial_waves()[:4]
                splines[id] = (phi_j, phit_j, nc, nct)
            phi_aj.append(phi_j)
            phit_aj.append(phit_j)
            nc_a.append([nc])
            nct_a.append([nct])
            if self.setups[a].data.has_corehole and not skip_core and \
                    self.nspins > 1:
                assert self.setups[a].data.lcorehole == 0
                work_setup = self.setups[a].data
                rmax = nc.get_cutoff()
                # work_setup.phicorehole_g
                phi_ch = work_setup.phicorehole_g
                phi_ch = np.where(abs(phi_ch) < 1e-160, 0, phi_ch)
                n_ch = np.dot(work_setup.fcorehole, phi_ch**2) / (4 * pi)
                # n_ch[0] = n_ch[1] # ch is allready taylored
                # fcorehole should divided by two - use scale for this
                n_ch_spl = work_setup.rgd.spline(n_ch, rmax, points=1000)
                n_ch_a.append([n_ch_spl])
                spos_ch.append(spos_ac[a])

        # Create localized functions from splines
        phi = BasisFunctions(gd, phi_aj)
        phit = BasisFunctions(gd, phit_aj)
        nc = LFC(gd, nc_a)
        nct = LFC(gd, nct_a)
        phi.set_positions(spos_ac)
        phit.set_positions(spos_ac)
        nc.set_positions(spos_ac)
        nct.set_positions(spos_ac)
        if spos_ch:
            nch = LFC(gd, n_ch_a)
            nch.set_positions(spos_ch)

        I_sa = np.zeros((self.nspins, len(spos_ac)))
        a_W = np.empty(len(phi.M_W), np.intc)
        W = 0
        for a in phi.atom_indices:
            nw = len(phi.sphere_a[a].M_w)
            a_W[W:W + nw] = a
            W += nw

        x_W = phi.create_displacement_arrays()[0]

        # We need the charges for the density matrices in order to add
        # nuclear charges at each atom.  Hence we use the aux partition:
        # The one where atoms are distributed according to which realspace
        # domain they belong to.
        D_asp = self.atomdist.to_aux(self.D_asp)

        rho_MM = np.zeros((phi.Mmax, phi.Mmax))
        for s, I_a in enumerate(I_sa):
            M1 = 0
            for a, setup in enumerate(self.setups):
                ni = setup.ni
                D_sp = D_asp.get(a)
                if D_sp is None:
                    D_sp = np.empty((self.nspins, ni * (ni + 1) // 2))
                else:
                    I_a[a] = ((setup.Nct) / self.nspins -
                              sqrt(4 * pi) *
                              np.dot(D_sp[s], setup.Delta_pL[:, 0]))

                    if not skip_core:
                        I_a[a] -= setup.Nc / self.nspins
                        if self.setups[a].data.has_corehole and \
                                self.nspins > 1:
                            I_a[a] += pow(-1, s) * \
                                self.setups[a].data.fcorehole / 2

                rank = D_asp.partition.rank_a[a]
                D_asp.partition.comm.broadcast(D_sp, rank)
                M2 = M1 + ni
                rho_MM[M1:M2, M1:M2] = unpack2(D_sp[s])
                M1 = M2

            assert np.all(n_sg[s].shape == phi.gd.n_c)
            phi.lfc.ae_valence_density_correction(rho_MM, n_sg[s], a_W, I_a,
                                                  x_W)
            phit.lfc.ae_valence_density_correction(-rho_MM, n_sg[s], a_W, I_a,
                                                   x_W)

        # wth is this?
        a_W = np.empty(len(nc.M_W), np.intc)
        W = 0
        for a in nc.atom_indices:
            nw = len(nc.sphere_a[a].M_w)
            a_W[W:W + nw] = a
            W += nw
        scale = 1.0 / self.nspins

        for s, I_a in enumerate(I_sa):

            if not skip_core:
                nc.lfc.ae_core_density_correction(scale, n_sg[s], a_W, I_a)
                # correct for ch here
                if spos_ch:
                    nch.lfc.ae_core_density_correction(- scale * pow(-1, s),
                                                       n_sg[s], a_W, I_a)
            nct.lfc.ae_core_density_correction(-scale, n_sg[s], a_W, I_a)
            D_asp.partition.comm.sum(I_a)
            N_c = gd.N_c
            g_ac = np.around(N_c * spos_ac).astype(int) % N_c - gd.beg_c

            if not skip_core:
                for I, g_c in zip(I_a, g_ac):
                    if (g_c >= 0).all() and (g_c < gd.n_c).all():
                        n_sg[s][tuple(g_c)] -= I / gd.dv

        return n_sg, gd

    def estimate_memory(self, mem):
        nspins = self.nspins
        nbytes = self.gd.bytecount()
        nfinebytes = self.finegd.bytecount()

        arrays = mem.subnode('Arrays')
        for name, size in [('nt_sG', nbytes * nspins),
                           ('nt_sg', nfinebytes * nspins),
                           ('nt_g', nfinebytes),
                           ('rhot_g', nfinebytes),
                           ('nct_G', nbytes)]:
            arrays.subnode(name, size)

        lfs = mem.subnode('Localized functions')
        for name, obj in [('nct', self.nct),
                          ('ghat', self.ghat)]:
            obj.estimate_memory(lfs.subnode(name))
        self.mixer.estimate_memory(mem.subnode('Mixer'), self.gd)

        # TODO
        # The implementation of interpolator memory use is not very
        # accurate; 20 MiB vs 13 MiB estimated in one example, probably
        # worse for parallel calculations.

    def get_spin_contamination(self, atoms, majority_spin=0):
        """Calculate the spin contamination.

        Spin contamination is defined as the integral over the
        spin density difference, where it is negative (i.e. the
        minority spin density is larger than the majority spin density.
        """

        if majority_spin == 0:
            smaj = 0
            smin = 1
        else:
            smaj = 1
            smin = 0
        nt_sg, gd = self.get_all_electron_density(atoms)
        dt_sg = nt_sg[smin] - nt_sg[smaj]
        dt_sg = np.where(dt_sg > 0, dt_sg, 0.0)
        return gd.integrate(dt_sg)

    def write(self, writer):
        writer.write(density=self.gd.collect(self.nt_xG) / Bohr**3,
                     atomic_density_matrices=pack_atomic_matrices(self.D_asp))

    def read(self, reader):
        nt_xG = self.gd.empty(self.ncomponents)
        self.gd.distribute(reader.density.density, nt_xG)
        nt_xG *= reader.bohr**3

        # Read atomic density matrices
        natoms = len(self.setups)
        atom_partition = AtomPartition(self.gd.comm, np.zeros(natoms, int),
                                       'density-gd')
        D_asp = self.setups.empty_atomic_matrix(self.ncomponents,
                                                atom_partition)
        self.atom_partition = atom_partition  # XXXXXX
        spos_ac = np.zeros((natoms, 3))  # XXXX
        self.atomdist = self.redistributor.get_atom_distributions(spos_ac)

        D_sP = reader.density.atomic_density_matrices
        if self.gd.comm.rank == 0:
            D_asp.update(unpack_atomic_matrices(D_sP, self.setups))
            D_asp.check_consistency()

        if self.collinear:
            nt_sG = nt_xG
            nt_vG = None
        else:
            nt_sG = nt_xG[:1]
            nt_vG = nt_xG[1:]

        self.initialize_directly_from_arrays(nt_sG, nt_vG, D_asp)

    def initialize_from_other_density(self, dens, kptband_comm):
        """Redistribute pseudo density and atomic density matrices.

        Collect dens.nt_sG and dens.D_asp to world master and distribute."""

        new_nt_sG = redistribute_array(dens.nt_sG, dens.gd, self.gd,
                                       self.nspins, kptband_comm)
        self.atom_partition, self.atomdist = \
            create_atom_partition_and_distibutions(self.gd, self.nspins,
                                                   self.setups,
                                                   self.redistributor,
                                                   kptband_comm)
        D_asp = \
            redistribute_atomic_matrices(dens.D_asp, self.gd, self.ncomponents,
                                         self.setups, self.atom_partition,
                                         kptband_comm)
        self.initialize_directly_from_arrays(new_nt_sG, None, D_asp)


class RealSpaceDensity(Density):
    def __init__(self, gd, finegd, nspins, collinear, charge, redistributor,
                 stencil=3,
                 background_charge=None):
        Density.__init__(self, gd, finegd, nspins, collinear,
                         charge, redistributor,
                         background_charge=background_charge)
        self.stencil = stencil

        self.interpolator = None

    def initialize(self, setups, timer, magmom_a, hund):
        Density.initialize(self, setups, timer, magmom_a, hund)

        # Interpolation function for the density:
        self.interpolator = Transformer(self.redistributor.aux_gd,
                                        self.finegd, self.stencil)

        spline_aj = []
        for setup in setups:
            if setup.nct is None:
                spline_aj.append([])
            else:
                spline_aj.append([setup.nct])
        self.nct = LFC(self.gd, spline_aj,
                       integral=[setup.Nct for setup in setups],
                       forces=True, cut=True)
        self.ghat = LFC(self.finegd, [setup.ghat_l for setup in setups],
                        integral=sqrt(4 * pi), forces=True)

    def set_positions(self, spos_ac, atom_partition):
        Density.set_positions(self, spos_ac, atom_partition)
        self.nct_G = self.gd.zeros()
        self.nct.add(self.nct_G, 1.0 / self.nspins)

    def interpolate_pseudo_density(self, comp_charge=None):
        """Interpolate pseudo density to fine grid."""
        if comp_charge is None:
            comp_charge, _Q_aL = self.calculate_multipole_moments()

        self.nt_sg = self.distribute_and_interpolate(self.nt_sG, self.nt_sg)

        # With periodic boundary conditions, the interpolation will
        # conserve the number of electrons.
        if not self.gd.pbc_c.all():
            # With zero-boundary conditions in one or more directions,
            # this is not the case.
            pseudo_charge = (self.background_charge.charge - self.charge -
                             comp_charge)
            if abs(pseudo_charge) > 1.0e-14:
                x = (pseudo_charge /
                     self.finegd.integrate(self.nt_sg).sum())
                self.nt_sg *= x

    def interpolate(self, in_xR, out_xR=None):
        """Interpolate array(s)."""

        # ndim will be 3 in finite-difference mode and 1 when working
        # with the AtomPAW class (spherical atoms and 1d grids)
        ndim = self.gd.ndim

        if out_xR is None:
            out_xR = self.finegd.empty(in_xR.shape[:-ndim])

        a_xR = in_xR.reshape((-1,) + in_xR.shape[-ndim:])
        b_xR = out_xR.reshape((-1,) + out_xR.shape[-ndim:])

        for in_R, out_R in zip(a_xR, b_xR):
            self.interpolator.apply(in_R, out_R)

        return out_xR

    def distribute_and_interpolate(self, in_xR, out_xR=None):
        in_xR = self.redistributor.distribute(in_xR)
        return self.interpolate(in_xR, out_xR)

    def calculate_pseudo_charge(self):
        self.nt_g = self.nt_sg.sum(axis=0)
        self.rhot_g = self.nt_g.copy()
        self.calculate_multipole_moments()
        self.ghat.add(self.rhot_g, self.Q_aL)
        self.background_charge.add_charge_to(self.rhot_g)

        if debug:
            charge = self.finegd.integrate(self.rhot_g) + self.charge
            if abs(charge) > self.charge_eps:
                raise RuntimeError('Charge not conserved: excess=%.9f' %
                                   charge)

    def get_pseudo_core_kinetic_energy_density_lfc(self):
        return LFC(self.gd,
                   [[setup.tauct] for setup in self.setups],
                   forces=True, cut=True)

    def calculate_dipole_moment(self):
        return self.finegd.calculate_dipole_moment(self.rhot_g)


def redistribute_array(nt_sG, gd1, gd2, nspins, kptband_comm):
    nt_sG = gd1.collect(nt_sG)
    new_nt_sG = gd2.empty(nspins)
    if kptband_comm.rank == 0:
        gd2.distribute(nt_sG, new_nt_sG)
    kptband_comm.broadcast(new_nt_sG, 0)
    return new_nt_sG


def create_atom_partition_and_distibutions(gd2, nspins, setups, redistributor,
                                           kptband_comm):
    natoms = len(setups)
    atom_partition = AtomPartition(gd2.comm, np.zeros(natoms, int),
                                   'density-gd')
    spos_ac = np.zeros((natoms, 3))  # XXXX
    atomdist = redistributor.get_atom_distributions(spos_ac)
    return atom_partition, atomdist


def redistribute_atomic_matrices(D_asp, gd2, nspins, setups, atom_partition,
                                 kptband_comm):
    D_sP = pack_atomic_matrices(D_asp)
    D_asp = setups.empty_atomic_matrix(nspins, atom_partition)
    if gd2.comm.rank == 0:
        if kptband_comm.rank > 0:
            nP = sum(setup.ni * (setup.ni + 1) // 2
                     for setup in setups)
            D_sP = np.empty((nspins, nP))
        kptband_comm.broadcast(D_sP, 0)
        D_asp.update(unpack_atomic_matrices(D_sP, setups))
        D_asp.check_consistency()
    return D_asp



# flake8: noqa
import numpy as np

from gpaw import debug
from gpaw.mpi import world
mpi_debug = lambda x, ordered=True: None  # silenced
from gpaw.overlap import Overlap
from gpaw.utilities import unpack
from gpaw.lfc import LocalizedFunctionsCollection as LFC


class PairOverlap:
    def __init__(self, gd, setups):
        self.gd = gd
        self.setups = setups
        self.ni_a = np.cumsum([0] + [setup.ni for setup in self.setups])

    def __len__(self):
        return self.ni_a[-1].item()

    def assign_atomic_pair_matrix(self, X_aa, a1, a2, dX_ii):
        X_aa[self.ni_a[a1]:self.ni_a[a1+1],\
             self.ni_a[a2]:self.ni_a[a2+1]] = dX_ii

    def extract_atomic_pair_matrix(self, X_aa, a1, a2):
        return X_aa[self.ni_a[a1]:self.ni_a[a1+1],\
                    self.ni_a[a2]:self.ni_a[a2+1]]

    def calculate_overlaps(self, spos_ac, lfc1, lfc2=None):
        raise RuntimeError('This is a virtual member function.')

    def calculate_atomic_pair_overlaps(
            self, lfs1, lfs2):  #XXX Move some code here from above...
        raise RuntimeError('This is a virtual member function.')


class GridPairOverlap(PairOverlap):
    def calculate_overlaps(self, spos_ac, lfc1, lfc2=None):
        # CONDITION: The two sets of splines must belong to the same kpoint!

        if lfc2 is None:
            lfc2 = lfc1

        if isinstance(lfc1, LFC) and isinstance(lfc2, LFC):
            return self.calculate_overlaps2(spos_ac, lfc1, lfc2)

        assert not isinstance(lfc1, LFC) and not isinstance(lfc2, LFC)

        nproj = len(self)
        X_aa = np.zeros((nproj, nproj), dtype=float)  # XXX always float?

        if debug:
            if world.rank == 0:
                print('DEBUG INFO')

            mpi_debug('lfc1.lfs_a.keys(): %s' % lfc1.lfs_a.keys())
            mpi_debug('lfc2.lfs_a.keys(): %s' % lfc2.lfs_a.keys())
            mpi_debug('N_c=%s, beg_c=%s, end_c=%s' %
                      (self.gd.N_c, self.gd.beg_c, self.gd.end_c))

            assert len(lfc1.spline_aj) == len(lfc1.spos_ac)  # not distributed
            assert len(lfc2.spline_aj) == len(lfc2.spos_ac)  # not distributed
            #assert lfc1.lfs_a.keys() == lfc2.lfs_a.keys() # XXX must they be equal?!?

        # Both loops are over all atoms in all domains
        for a1, spline1_j in enumerate(lfc1.spline_aj):
            # We assume that all functions have the same cut-off:
            rcut1 = spline1_j[0].get_cutoff()
            if debug:
                mpi_debug('a1=%d, spos1_c=%s, rcut1=%g, ni1=%d' %
                          (a1, spos_ac[a1], rcut1, self.setups[a1].ni))

            for a2, spline2_j in enumerate(lfc2.spline_aj):
                # We assume that all functions have the same cut-off:
                rcut2 = spline2_j[0].get_cutoff()
                if debug:
                    mpi_debug('  a2=%d, spos2_c=%s, rcut2=%g, ni2=%d' %
                              (a2, spos_ac[a2], rcut2, self.setups[a2].ni))

                X_ii = self.extract_atomic_pair_matrix(X_aa, a1, a2)

                b1 = 0
                for beg1_c, end1_c, sdisp1_c in self.gd.get_boxes(
                        spos_ac[a1], rcut1,
                        cut=False):  # loop over lfs1.box_b instead?
                    if debug:
                        mpi_debug(
                            '    b1=%d, beg1_c=%s, end1_c=%s, sdisp1_c=%s' %
                            (b1, beg1_c, end1_c, sdisp1_c),
                            ordered=False)

                    # Atom a1 has at least one piece so the LFC has LocFuncs
                    lfs1 = lfc1.lfs_a[a1]

                    # Similarly, the LocFuncs must have the piece at hand
                    box1 = lfs1.box_b[b1]

                    if debug:
                        assert lfs1.dtype == lfc1.dtype
                        assert self.setups[
                            a1].ni == lfs1.ni, 'setups[%d].ni=%d, lfc1.lfs_a[%d].ni=%d' % (
                                a1, self.setups[a1].ni, a1, lfs1.i)

                    b2 = 0
                    for beg2_c, end2_c, sdisp2_c in self.gd.get_boxes(
                            spos_ac[a2], rcut2,
                            cut=False):  # loop over lfs2.box_b instead?
                        if debug:
                            mpi_debug(
                                '      b2=%d, beg2_c=%s, end2_c=%s, sdisp2_c=%s'
                                % (b2, beg2_c, end2_c, sdisp2_c),
                                ordered=False)

                        # Atom a2 has at least one piece so the LFC has LocFuncs
                        lfs2 = lfc2.lfs_a[a2]

                        # Similarly, the LocFuncs must have the piece at hand
                        box2 = lfs2.box_b[b2]

                        if debug:
                            assert lfs2.dtype == lfc2.dtype
                            assert self.setups[
                                a2].ni == lfs2.ni, 'setups[%d].ni=%d, lfc2.lfs_a[%d].ni=%d' % (
                                    a2, self.setups[a2].ni, a2, lfs2.ni)

                        # Find the intersection of the two boxes
                        beg_c = np.max((beg1_c, beg2_c), axis=0)
                        end_c = np.min((end1_c, end2_c), axis=0)

                        if debug:
                            mpi_debug('        beg_c=%s, end_c=%s, size_c=%s' %
                                      (beg_c, end_c, tuple(end_c - beg_c)),
                                      ordered=False)

                        # Intersection is non-empty, add overlap contribution
                        if (beg_c < end_c).all():
                            bra_iB1 = box1.get_functions()
                            w1slice = [slice(None)]+[slice(b,e) for b,e in \
                                zip(beg_c-beg1_c, end_c-beg1_c)]

                            ket_iB2 = box2.get_functions()
                            w2slice = [slice(None)]+[slice(b,e) for b,e in \
                                zip(beg_c-beg2_c, end_c-beg2_c)]

                            X_ii += self.gd.dv * np.inner( \
                                bra_iB1[w1slice].reshape((lfs1.ni,-1)), \
                                ket_iB2[w2slice].reshape((lfs2.ni,-1))) #XXX phase factors for kpoints

                            del bra_iB1, ket_iB2

                        b2 += 1

                    b1 += 1

        self.gd.comm.sum(X_aa)  # better to sum over X_ii?
        return X_aa

    def calculate_overlaps2(self, spos_ac, lfc1, lfc2=None):
        # CONDITION: The two sets of splines must belong to the same kpoint!

        if lfc2 is None:
            lfc2 = lfc1

        assert isinstance(lfc1, LFC) and isinstance(lfc2, LFC)

        nproj = len(self)
        X_aa = np.zeros((nproj, nproj), dtype=float)  # XXX always float?

        if debug:
            if world.rank == 0:
                print('DEBUG INFO')

            mpi_debug('len(lfc1.sphere_a): %d, lfc1.atom_indices: %s' %
                      (len(lfc1.sphere_a), lfc1.atom_indices))
            mpi_debug('len(lfc2.sphere_a): %d, lfc2.atom_indices: %s' %
                      (len(lfc2.sphere_a), lfc2.atom_indices))
            mpi_debug('N_c=%s, beg_c=%s, end_c=%s' %
                      (self.gd.N_c, self.gd.beg_c, self.gd.end_c))

        if debug:
            assert len(lfc1.sphere_a) == len(
                lfc2.sphere_a)  # XXX must they be equal?!?

        # Both a-loops are over all relevant atoms which affect this domain
        for a1 in lfc1.atom_indices:
            sphere1 = lfc1.sphere_a[a1]

            # We assume that all functions have the same cut-off:
            spline1_j = sphere1.spline_j
            rcut1 = spline1_j[0].get_cutoff()
            if debug:
                mpi_debug('a1=%d, spos1_c=%s, rcut1=%g, ni1=%d' %
                          (a1, spos_ac[a1], rcut1, self.setups[a1].ni),
                          ordered=False)

            for a2 in lfc2.atom_indices:
                sphere2 = lfc2.sphere_a[a2]

                # We assume that all functions have the same cut-off:
                spline2_j = sphere2.spline_j
                rcut2 = spline2_j[0].get_cutoff()
                if debug:
                    mpi_debug('  a2=%d, spos2_c=%s, rcut2=%g, ni2=%d' %
                              (a2, spos_ac[a2], rcut2, self.setups[a2].ni),
                              ordered=False)

                X_ii = self.extract_atomic_pair_matrix(X_aa, a1, a2)

                b1 = 0
                for beg1_c, end1_c, sdisp1_c in self.gd.get_boxes(
                        spos_ac[a1], rcut1,
                        cut=False):  # loop over lfs1.box_b instead?
                    if debug:
                        mpi_debug(
                            '    b1=%d, beg1_c=%s, end1_c=%s, sdisp1_c=%s' %
                            (b1, beg1_c, end1_c, sdisp1_c),
                            ordered=False)

                    b2 = 0
                    for beg2_c, end2_c, sdisp2_c in self.gd.get_boxes(
                            spos_ac[a2], rcut2,
                            cut=False):  # loop over lfs2.box_b instead?
                        if debug:
                            mpi_debug(
                                '      b2=%d, beg2_c=%s, end2_c=%s, sdisp2_c=%s'
                                % (b2, beg2_c, end2_c, sdisp2_c),
                                ordered=False)

                        # Find the intersection of the two boxes
                        beg_c = np.max((beg1_c, beg2_c), axis=0)
                        end_c = np.min((end1_c, end2_c), axis=0)

                        if debug:
                            mpi_debug('        beg_c=%s, end_c=%s, size_c=%s' %
                                      (beg_c, end_c, tuple(end_c - beg_c)),
                                      ordered=False)

                        # Intersection is non-empty, add overlap contribution
                        if (beg_c < end_c).all():
                            i1 = 0
                            for j1, spline1 in enumerate(spline1_j):
                                bra1_mB = spline1.get_functions(self.gd, \
                                    beg_c, end_c, spos_ac[a1]-sdisp1_c)
                                nm1 = bra1_mB.shape[0]

                                i2 = 0
                                for j2, spline2 in enumerate(spline2_j):
                                    ket2_mB = spline2.get_functions(self.gd, \
                                        beg_c, end_c, spos_ac[a2]-sdisp2_c)
                                    nm2 = ket2_mB.shape[0]

                                    X_mm = X_ii[i1:i1 + nm1, i2:i2 + nm2]
                                    X_mm += self.gd.dv * np.inner( \
                                        bra1_mB.reshape((nm1,-1)), \
                                        ket2_mB.reshape((nm2,-1))) #XXX phase factors for kpoints

                                    del ket2_mB
                                    i2 += nm2

                                del bra1_mB
                                i1 += nm1

                        b2 += 1

                    b1 += 1

        self.gd.comm.sum(X_aa)  # better to sum over X_ii?
        return X_aa


class ProjectorPairOverlap(Overlap, GridPairOverlap):
    """
    TODO
    """
    def __init__(self, wfs, atoms):
        """TODO

        Attributes:

        ============  ======================================================
        ``B_aa``      < p_i^a | p_i'^a' >
        ``xO_aa``     TODO
        ``dC_aa``     TODO
        ``xC_aa``     TODO
        ============  ======================================================
        """

        Overlap.__init__(self, wfs.orthoksl, wfs.timer)
        GridPairOverlap.__init__(self, wfs.gd, wfs.setups)
        self.natoms = len(atoms)
        if debug:
            assert len(self.setups) == self.natoms
        self.update(wfs, atoms)

    def update(self, wfs, atoms):
        self.timer.start('Update two-center overlap')

        nproj = len(self)
        """
        self.B_aa = np.zeros((nproj, nproj), dtype=float) #always float?
        for a1,setup1 in enumerate(self.setups):
            for a2 in wfs.pt.my_atom_indices:
                setup2 = self.setups[a2]
                R = (atoms[a1].get_position() - atoms[a2].get_position()) / Bohr

                if a1 == a2:
                    B_ii = setup1.B_ii
                else:
                    B_ii = projector_overlap_matrix2(setup1, setup2, R)
                #if a1 < a2:
                #    B_ii = projector_overlap_matrix2(setup1, setup2, R)
                #elif a1 == a2:
                #    B_ii = setup1.B_ii
                #else:
                #    B_ii = self.B_aa[ni_a[a2]:ni_a[a2+1], ni_a[a1]:ni_a[a1+1]].T

                #self.B_aa[self.ni_a[a1]:self.ni_a[a1+1], \
                #          self.ni_a[a2]:self.ni_a[a2+1]] = B_ii
                self.assign_atomic_pair_matrix(self.B_aa, a1, a2, B_ii)
        self.gd.comm.sum(self.B_aa) #TODO too heavy?
        """
        #self.B_aa = overlap_projectors(wfs.gd, wfs.pt, wfs.setups)

        self.B_aa = self.calculate_overlaps(wfs.spos_ac, wfs.pt)

        # Create two-center (block-diagonal) coefficients for overlap operator
        dO_aa = np.zeros((nproj, nproj), dtype=float)  #always float?
        for a, setup in enumerate(self.setups):
            self.assign_atomic_pair_matrix(dO_aa, a, a, setup.dO_ii)

        # Calculate two-center rotation matrix for overlap projections
        self.xO_aa = self.get_rotated_coefficients(dO_aa)

        # Calculate two-center coefficients for inverse overlap operator
        lhs_aa = np.eye(nproj) + self.xO_aa
        rhs_aa = -dO_aa
        self.dC_aa = np.linalg.solve(lhs_aa.T, rhs_aa.T).T  #TODO parallel

        # Calculate two-center rotation matrix for inverse overlap projections
        self.xC_aa = self.get_rotated_coefficients(self.dC_aa)

        self.timer.stop('Update two-center overlap')

    def get_rotated_coefficients(self, X_aa):
        r"""Rotate two-center projector expansion coefficients with
        the projector-projector overlap integrals as basis.

        Performs the following operation and returns the result::

                    ---
            a1,a3   \       a1    a2     a2,a3
           Y      =  )  <  p   | p   >  X
            i1,i3   /       i1    i2     i2,i3
                    ---
                   a2,i2
        """
        return np.dot(self.B_aa, X_aa)

    def apply_to_atomic_matrices(self, dI_asp, P_axi, wfs, kpt, shape=()):

        self.timer.start('Update two-center projections')

        nproj = len(self)
        dI_aa = np.zeros((nproj, nproj), dtype=float)  #always float?

        for a, dI_sp in dI_asp.items():
            dI_p = dI_sp[kpt.s]
            dI_ii = unpack(dI_p)
            self.assign_atomic_pair_matrix(dI_aa, a, a, dI_ii)
        self.gd.comm.sum(dI_aa)  #TODO too heavy?

        dM_aa = self.get_rotated_coefficients(dI_aa)
        Q_axi = wfs.pt.dict(shape, zero=True)
        for a1 in range(self.natoms):
            if a1 in Q_axi.keys():
                Q_xi = Q_axi[a1]
            else:
                # Atom a1 is not in domain so allocate a temporary buffer
                Q_xi = np.zeros(shape + (self.setups[a1].ni, ),
                                dtype=wfs.pt.dtype)  #TODO
            for a2, P_xi in P_axi.items():
                dM_ii = self.extract_atomic_pair_matrix(dM_aa, a1, a2)
                Q_xi += np.dot(P_xi, dM_ii.T)  #sum over a2 and last i in dM_ii
            self.gd.comm.sum(Q_xi)

        self.timer.stop('Update two-center projections')

        return Q_axi

    def apply(self,
              a_xG,
              b_xG,
              wfs,
              kpt,
              calculate_P_ani=True,
              extrapolate_P_ani=False):
        """Apply the overlap operator to a set of vectors.

        Parameters
        ==========
        a_nG: ndarray
            Set of vectors to which the overlap operator is applied.
        b_nG: ndarray, output
            Resulting S times a_nG vectors.
        kpt: KPoint object
            k-point object defined in kpoint.py.
        calculate_P_ani: bool
            When True, the integrals of projector times vectors
            P_ni = <p_i | a_nG> are calculated.
            When False, existing P_ani are used
        extrapolate_P_ani: bool
            When True, the integrals of projector times vectors#XXX TODO
            P_ni = <p_i | a_nG> are calculated.
            When False, existing P_ani are used

        """

        self.timer.start('Apply overlap')
        b_xG[:] = a_xG
        shape = a_xG.shape[:-3]
        P_axi = wfs.pt.dict(shape)

        if calculate_P_ani:
            wfs.pt.integrate(a_xG, P_axi, kpt.q)
        else:
            for a, P_ni in kpt.P_ani.items():
                P_axi[a][:] = P_ni

        Q_axi = wfs.pt.dict(shape)
        for a, Q_xi in Q_axi.items():
            Q_xi[:] = np.dot(P_axi[a], self.setups[a].dO_ii)

        wfs.pt.add(b_xG, Q_axi, kpt.q)
        self.timer.stop('Apply overlap')

        if extrapolate_P_ani:
            for a1 in range(self.natoms):
                if a1 in Q_axi.keys():
                    Q_xi = Q_axi[a1]
                    Q_xi[:] = P_axi[a1]
                else:
                    # Atom a1 is not in domain so allocate a temporary buffer
                    Q_xi = np.zeros(shape + (self.setups[a1].ni, ),
                                    dtype=wfs.pt.dtype)  #TODO
                for a2, P_xi in P_axi.items():
                    # xO_aa are the overlap extrapolators across atomic pairs
                    xO_ii = self.extract_atomic_pair_matrix(self.xO_aa, a1, a2)
                    Q_xi += np.dot(P_xi,
                                   xO_ii.T)  #sum over a2 and last i in xO_ii
                self.gd.comm.sum(Q_xi)

            return Q_axi
        else:
            return P_axi

    def apply_inverse(self,
                      a_xG,
                      b_xG,
                      wfs,
                      kpt,
                      calculate_P_ani=True,
                      extrapolate_P_ani=False):

        self.timer.start('Apply inverse overlap')
        b_xG[:] = a_xG
        shape = a_xG.shape[:-3]
        P_axi = wfs.pt.dict(shape)

        if calculate_P_ani:
            wfs.pt.integrate(a_xG, P_axi, kpt.q)
        else:
            for a, P_ni in kpt.P_ani.items():
                P_axi[a][:] = P_ni

        Q_axi = wfs.pt.dict(shape, zero=True)
        for a1 in range(self.natoms):
            if a1 in Q_axi.keys():
                Q_xi = Q_axi[a1]
            else:
                # Atom a1 is not in domain so allocate a temporary buffer
                Q_xi = np.zeros(shape + (self.setups[a1].ni, ),
                                dtype=wfs.pt.dtype)  #TODO
            for a2, P_xi in P_axi.items():
                # dC_aa are the inverse coefficients across atomic pairs
                dC_ii = self.extract_atomic_pair_matrix(self.dC_aa, a1, a2)
                Q_xi += np.dot(P_xi, dC_ii.T)  #sum over a2 and last i in dC_ii
            self.gd.comm.sum(Q_xi)

        wfs.pt.add(b_xG, Q_axi, kpt.q)
        self.timer.stop('Apply inverse overlap')

        if extrapolate_P_ani:
            for a1 in range(self.natoms):
                if a1 in Q_axi.keys():
                    Q_xi = Q_axi[a1]
                    Q_xi[:] = P_axi[a1]
                else:
                    # Atom a1 is not in domain so allocate a temporary buffer
                    Q_xi = np.zeros(shape + (self.setups[a1].ni, ),
                                    dtype=wfs.pt.dtype)  #TODO
                for a2, P_xi in P_axi.items():
                    # xC_aa are the inverse extrapolators across atomic pairs
                    xC_ii = self.extract_atomic_pair_matrix(self.xC_aa, a1, a2)
                    Q_xi += np.dot(P_xi,
                                   xC_ii.T)  #sum over a2 and last i in xC_ii
                self.gd.comm.sum(Q_xi)

            return Q_axi
        else:
            return P_axi



import pytest
from gpaw.utilities import compiled_with_libvdwxc

from ase.build import bulk

from gpaw import GPAW, Davidson, Mixer, PW
from gpaw.xc.libvdwxc import vdw_mbeef

from gpaw.test import gen

pytestmark = pytest.mark.skipif(not compiled_with_libvdwxc(),
                                reason='not compiled_with_libvdwxc()')


@pytest.mark.mgga
def test_vdw_libvdwxc_mbeef():
    setup = gen('Si', xcname='PBEsol')

    system = bulk('Si')
    calc = GPAW(mode=PW(200), xc=vdw_mbeef(),
                kpts=(2, 2, 2),
                nbands=4,
                convergence=dict(density=1e-6),
                mixer=Mixer(1.0),
                eigensolver=Davidson(4),
                setups={'Si': setup})
    system.calc = calc
    e = system.get_potential_energy()
    ref = -60.53662105617721
    err = abs(e - ref)
    print('e=%r ref=%r err=%r' % (e, ref, err))
    # It would be reasonable to put 1e-6 as tolerance,
    # but the value changes by 4e-4 depending on libxc version.
    # See https://gitlab.com/gpaw/gpaw/issues/161 .
    assert err < 1e-3, err



"""Test Tran Blaha potential."""
import pytest
from ase.dft.bandgap import bandgap
from ase.build import bulk
from gpaw import GPAW, PW, Mixer


@pytest.mark.later
@pytest.mark.libxc
@pytest.mark.mgga
def test_xc_tb09(in_tmp_dir):
    def xc(name):
        return {'name': name, 'stencil': 1}

    k = 8
    atoms = bulk('Si')
    atoms.calc = GPAW(mode=PW(300),
                      mixer=Mixer(0.8, 10, 50.0),
                      kpts={'size': (k, k, k), 'gamma': True},
                      xc=xc('TB09'),
                      convergence={'bands': -3},
                      txt='si.txt')
    atoms.get_potential_energy()
    gap, (sv, kv, nv), (sc, kc, nc) = bandgap(atoms.calc)
    c = atoms.calc.hamiltonian.xc.c
    print(gap, kv, kc)
    print('c:', c)
    assert abs(gap - 1.246) < 0.01
    assert kv == 0 and kc == 12
    assert abs(c - 1.135) < 0.01
